
#%%
#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op

# pip
import xarray as xr
import numpy as np
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
import pickle
from datetime import timedelta, datetime
import numpy.matlib as nm
from cftime._cftime import DatetimeGregorian


# DEV: override installed teslakit
import sys
sys.path.insert(0, op.join(os.path.abspath(''), '..', '..'))

# teslakit
from teslakit.database import Database
from teslakit.io.matlab import ReadMatfile
from teslakit.rbf import RBF_Interpolation
from teslakit.util.time_operations import date2datenum as d2d
from teslakit.plotting.extremes import Plot_ReturnPeriodValidation
from teslakit.util.time_operations import DateConverter_Mat2Py


## Database and Site parameters
# %%
# --------------------------------------
# Teslakit database

p_data = r'/Users/anacrueda/Documents/Proyectos/TESLA/Teslakit/data'
db = Database(p_data)

# set site
db.SetSite('KWAJALEIN')

# %%
# --------------------------------------
# load data and set parameters
#  wave data reconstructed at 30m water depth

p_sim = r'/Users/anacrueda/Documents/Proyectos/SERDP/nearshores_waves_albaR/'

p_dataset = op.join(p_sim, 'sim10_propagg21.pkl')
fileObject = open(p_dataset,'rb')
subset = pickle.load(fileObject)

p_datasetO = op.join(p_sim, 'complete_h_offshore.nc')
dataO = xr.open_dataset(p_datasetO)

p_dataO_storm = op.join(p_sim, 'SIM_WAVES_TCs_offshore.nc')
data_storm = xr.open_dataset(p_dataO_storm)


# %%
#define reef characteristics
Rslope = 0.0505*np.ones(len(data_storm.time))
Bslope = 0.1667*np.ones(len(data_storm.time))
Rw = 250*np.ones(len(data_storm.time))
Cf = 0.0105*np.ones(len(data_storm.time))


# %% time vector for the hourly data

fechas_nuevas = [datetime(1700, 1, 1, 00) + timedelta(hours=x) for x in range(int(len(dataO.time)))]
#dataO_mod = dataO.loc[~dataO.time.duplicated(keep='first')]

#print(len(fechas_nuevas))

dataO['time'] = fechas_nuevas
print(dataO.time)

#print(dataO.time[-1].values)
#ttt = dataO.time.values[:]
#print(len(data_storm_time_h_rand))
#print(data_storm_time_h_rand[-1] in ttt)


# %%
# NICO: DE ESTA SELECCION COGEMOS MMSL Y SS
#dataO_storm = dataO.sel(n_sim=0, time=data_storm.time, drop=True)

# NICO: NOS CREAMOS UN VECTOR DE TIEMPO IGUAL QUE DATA_STORM.TIME PERO CON HORA
# ALEATORIA

#data_storm_time_h_rand = [

 #   d2d(d) + timedelta(hours=np.random.randint(23)) for d in data_storm.time.values[:]

#]
#print(data_storm_time_h_rand)

# NICO: COGEMOS AT DE AQUI
#el ultimo valor no puede ser random porque la serie sintetica acaba ese dia a las 00
# OJO chapuza! nico help!!
#data_storm_time_h_rand[-1] = datetime(2700,1,1,00)

#AT = dataO.sel(n_sim=0,time=data_storm_time_h_rand[:]).AT
#dataO_storm = dataO.sel(n_sim=0, time=data_storm.time)

#Level = dataO_storm.SS + dataO_storm.MMSL + AT

# %% varying AT, SS and MMSL for each simulation

Level_xds = xr.Dataset({})
for i_sim in dataO.n_sim:
    dataO_storm = dataO.sel(n_sim=i_sim, time=data_storm.time, drop=True)
    data_storm_time_h_rand = [

        d2d(d) + timedelta(hours=np.random.randint(23)) for d in data_storm.time.values[:]

    ]
    # el ultimo valor no puede ser random porque la serie sintetica acaba ese dia a las 00
    # OJO chapuza! nico help!!
    data_storm_time_h_rand[-1] = datetime(2700, 1, 1, 00)
    AT = dataO.sel(n_sim=i_sim, time=data_storm_time_h_rand[:]).AT

    Level_s = dataO_storm.SS + dataO_storm.MMSL + AT

    Level = xr.Dataset({
        'level': ('time', xr.DataArray(Level_s)),
    },
        coords={'time': data_storm.time}
    )
    #print(Level)
    #print(i_sim)
    Level_xds = xr.concat([Level_xds, Level], dim='n_sim')

# %%
# dataset with wave and reef characteristics
print(np.shape(subset['Hs']))

Hs = np.reshape([subset['Hs']], (10,len(data_storm.time)),'C')
Tp = np.reshape([subset['Tp']], (10,len(data_storm.time)),'C')
Dir = np.reshape([subset['Dir']], (10,len(data_storm.time)),'C')

# Level_r = nm.repmat(Level, 10, 1)
#Level_r = np.reshape([Level_xds['level']], (10,len(data_storm.time)),'C')


sim_wvs = xr.Dataset(
    {'Hs': (('n_sim', 'time'), Hs),
     'Tp': (('n_sim', 'time'), Tp),
     'Dir': (('n_sim', 'time'), Dir),
     'Level': (('n_sim', 'time'), Level_xds['level'])
},
    coords={'n_sim': dataO.n_sim, 'time': data_storm.time}
)

print(sim_wvs)


# %%

hs_lo2 = sim_wvs['Hs']/(1.5613*sim_wvs['Tp']**2)

print(hs_lo2)

xds_subset = xr.Dataset(
    {
        'hs': (('n_sim','time'), sim_wvs['Hs']), #subset['Hs']
        'tp': (('n_sim','time'),sim_wvs['Tp']),
        'dir': (('n_sim','time'),sim_wvs['Dir']),
        'level': (('n_sim','time'),sim_wvs['Level']),
        'hs_lo2': (('n_sim','time'),hs_lo2),
        'rslope': Rslope,
        'bslope': Bslope,
        'rwidth': Rw,
        'cf': Cf,
    },

    coords={'n_sim': dataO.n_sim, 'time': data_storm.time},
    # dims = {'n_sim':10,'time':len(time)}

)

print(xds_subset)

# load min, max, of simulated reef and wave conditions and rbfs coefficients

p_rbf = r'/Users/anacrueda/Documents/Proyectos/SERDP/HyCreWWcode/RBF_coefficients/'

p_min = op.join(p_rbf, 'Min_from_simulations.mat')
p_max = op.join(p_rbf, 'Max_from_simulations.mat')

smin = ReadMatfile(p_min)
smax = ReadMatfile(p_max)

# data max and min
dl = np.column_stack([smin['minimum'], smax['maximum']])

print(dl)

d = {
    'hs': dl[1],
    'tp': dl[2],
    'level':dl[0],
    'rslope': dl[3],
    'bslope': dl[4],
    'rwidth': dl[5],
    'cf': dl[6],
    'hs_lo2': [0.005, 0.05],
}

for k in d.keys():
    print(k, d[k])

# %%
# input data quality check
#['level','hs','tp','rslope','bslope','rwidth','cf']:

runup_all_xds = xr.Dataset({})

for i_sim in dataO.n_sim:
    xds_f = xds_subset.sel(n_sim=i_sim)

    for vn in xds_f.variables:
        if vn in d.keys():
            # get limits from dictionary
            l, h = tuple(d[vn])

            # find positions inside limits
            xds_f[vn] = xds_f[vn].where((xds_f[vn] >= l) & (xds_f[vn] <= h))

    print('xds_f')
    print(xds_f)
    print()
    # Normalize reef and water level parameters data
    xds_reef = xr.Dataset({
       'level': (xds_f['level']),
       'rslope': (xds_f['rslope']),
       'bslope': (xds_f['bslope']),
       'rwidth': (xds_f['rwidth']),
       'cf': (xds_f['cf']),
        },
         coords={'time': xds_f['time']}
        )

    print('xds_reef')
    print(xds_reef)
    print()
    xds_n = xds_reef.copy()
    for k in xds_reef.variables:
        if k in d.keys():
           v = xds_n[k].values[:]
           l, h = tuple(d[k])
           xds_n[k] =(('time',), (v-l)/(h-l))
    print('xds_n')
    print(xds_n)

    # Hycreeww
    # 15 different wave conditions were considered for each water level and reef characteristics
    # the runup is calculated and then it is linearly interpolated for the actual wave conditions.
    ncases = 15
    runup = []
    # Build RBF target numpy array
    target = np.column_stack(
         (xds_n['level'], xds_n['rslope'], xds_n['bslope'], xds_n['rwidth'], xds_n['cf'])
    )
    target_wavecon = np.column_stack(
      (xds_f['hs'],xds_f['tp'],xds_f['hs_lo2'])
    )

    f_tot = []
    for j in range(0, ncases):
        p_coeff= op.join(p_rbf, 'Coeffs_Runup_Xbeach_test' + str(j+1) + '.mat')
        coeff = ReadMatfile(p_coeff)
        rbf_constant = coeff['coeffsave']['RBFConstant']
        rbf_coeff = coeff['coeffsave']['rbfcoeff']
        nodes = coeff['coeffsave']['x']
        x = target
        f_temp = RBF_Interpolation(rbf_constant, rbf_coeff, nodes,x.T)
        f_tot.append(f_temp)
    print(f_tot)

    #linear interpolation
    hs=[1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5] #Wave heigth conditions used in RBF
    hs_lo=[0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05 ] #Wave conditions used in RBF

    #print(np.shape(f_tot))
    #print(target_wavecon.shape)
    to_z = np.array(f_tot)
    #print(to_z.shape)
    RU = []
    for j in range (0,len(target)):
        x=hs; y=hs_lo; z=to_z[:,j]
        hs_e=target_wavecon[j,0]
        #tp_e=target_wavecon[j,1]
        hs_lo_e=target_wavecon[j,2]
        vq = griddata((x,y),z,(hs_e,hs_lo_e),method='linear')
        RU.append(vq)
    print(RU)
    print()
    #print(i_sim)

    runup_x = np.array(RU)

    runup_xds = xr.Dataset({
        'runup': ('time', xr.DataArray(runup_x)),
    },
        coords={'time': data_storm.time}
    )
    runup_all_xds = xr.concat([runup_all_xds, runup_xds], dim='n_sim')

# %%

true_runup = runup_all_xds['runup'] + xds_subset['level']

# we need to add the level to the hycreww estimation of Ru to get the proper runup
xds_results = xr.Dataset({
        'hs': (('n_sim','time'), xds_subset['hs']),
        'tp': (('n_sim','time'), xds_subset['tp']),
        'dir': (('n_sim','time'), xds_subset['dir']),
        'level': (('n_sim','time'), xds_subset['level']),
        'hs_lo2': (('n_sim','time'), xds_subset['hs_lo2']),
        'rslope': (xds_subset['rslope']),
        'bslope': (xds_subset['bslope']),
        'rwidth': (xds_subset['rwidth']),
        'cf': (xds_subset['cf']),
        'runup': (('n_sim','time'), true_runup),
        },

         coords={'n_sim': dataO.n_sim, 'time': data_storm.time},

    )

print(xds_results)

#%%
file_Name = r'/Users/anacrueda/Documents/Proyectos/SERDP/nearshores_waves_albaR/runup_all_sim_test_hycreww_py.pkl'
# open the file for writing
fileO = open(file_Name,'wb')

# this writes the object a to the
# file named 'testfile'
pickle.dump(xds_results,fileO)

# here we close the fileObject
fileO.close()


#print(xds_results)

#runup_nan = xds_results.where(np.isnan(xds_results.runup) == 'True', drop='True')

#print(runup_nan)

# %% figure annual maxima

# load historical
file_hist =r'/Users/anacrueda/Documents/Proyectos/SERDP/nearshores_waves_albaR/runup_hist_hycreww_py.pkl'

df = open(file_hist, 'rb')
data_hist = pickle.load(df)

print(data_hist)

# load simulated
file_sim= r'/Users/anacrueda/Documents/Proyectos/SERDP/nearshores_waves_albaR/runup_all_sim_test_hycreww_py.pkl'

df_sim = open(file_sim, 'rb')
data_sim= pickle.load(df_sim)

hist_A = data_hist['runup'].groupby('time.year').max(dim='time')
sim_A = data_sim['runup'].groupby('time.year').max(dim='time')

# Return Period historical vs. simulations
Plot_ReturnPeriodValidation(hist_A, sim_A)



from datetime import timedelta
from typing import Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from numpy import polyfit

from ..config.paths import get_paths
from ..core.constants import EARTH_RADIUS
from ..core.geo import geodesic_distance, geodesic_distance_azimuth, shoot

PATHS = get_paths(verbose=True)

# Configuration dictionaries and constants for IBTrACS data
centers_config_params: Dict[str, Dict[str, Union[str, List[str], float]]] = {
    "USA": {
        "id": "usa",
        "basins": ["NA", "SA", "EP", "WP", "SP", "NI", "SI"],
        "windfac": 1,
        "color": "r",
    },
    "TOKYO": {
        "id": "tokyo",
        "basins": ["EP", "WP"],
        "windfac": 0.93,
        "color": "lightcoral",
    },
    "CMA": {
        "id": "cma",
        "basins": ["EP", "WP"],
        "windfac": 0.99,
        "color": "lime",
    },
    "HKO": {
        "id": "hko",
        "basins": ["EP", "WP"],
        "windfac": 0.93,
        "color": "gold",
    },
    "NEWDELHI": {
        "id": "newdelhi",
        "basins": ["WP", "NI"],
        "windfac": 0.99,
        "color": "m",
    },
    "REUNION": {
        "id": "reunion",
        "basins": ["SI"],
        "windfac": 0.93,
        "color": "magenta",
    },
    "BOM": {
        "id": "bom",
        "basins": ["SP", "SI"],
        "windfac": 0.93,
        "color": "b",
    },
    "WELLINGTON": {
        "id": "wellington",
        "basins": ["SP", "SI"],
        "windfac": 0.93,
        "color": "slategrey",
    },
    "NADI": {
        "id": "nadi",
        "basins": ["SP"],
        "windfac": 0.93,
        "color": "c",
    },
    "WMO": {
        "id": "wmo",
        "basins": ["NA", "SA", "EP", "WP", "SP", "NI", "SI"],
        "windfac": 1,
        "color": "k",
    },
    "FORECAST": {
        "id": "forecast",
        "basins": ["AL", "LS", "EP", "CP", "WP", "IO", "SH"],
        "windfac": 1,
        "color": "k",
    },
}

all_centers: List[str] = [
    "USA",
    "TOKYO",
    "CMA",
    "HKO",
    "NEWDELHI",
    "REUNION",
    "BOM",
    "WELLINGTON",
    "NADI",
    "WMO",
]

all_basins: List[str] = ["NA", "SA", "EP", "WP", "SP", "NI", "SI"]


def get_center_information(center: str = "WMO") -> Dict[str, Union[str, None]]:
    """
    Get the center information from the configuration parameters for IBTrACS data.

    Parameters
    ----------
    center : str, optional
        The name of the center to get information for. Default is "WMO".

    Returns
    -------
    Dict[str, Union[str, None]]
        A dictionary containing the center information with the following keys:
        - source: Center ID
        - basin: Basin identifier
        - time: Time variable name
        - longitude: Longitude variable name
        - latitude: Latitude variable name
        - pressure: Pressure variable name
        - maxwinds: Maximum winds variable name
        - rmw: Radius of maximum winds variable name (if available)
        - dist2land: Distance to land variable name

    Raises
    ------
    ValueError
        If the specified center is not found in the configuration parameters.

    Examples
    --------
    >>> info = get_center_information("USA")
    >>> info["source"]
    'usa'
    >>> info["rmw"]
    'usa_rmw'
    """

    if center not in centers_config_params:
        raise ValueError(
            f"Center {center} not found. Available centers: {list(centers_config_params.keys())}"
        )

    return {
        "source": centers_config_params[center]["id"],
        "basin": "basin",
        "time": "time",
        "longitude": "lon"
        if center == "WMO"
        else centers_config_params[center]["id"] + "_lon",
        "latitude": "lat"
        if center == "WMO"
        else centers_config_params[center]["id"] + "_lat",
        "pressure": centers_config_params[center]["id"] + "_pres",
        "maxwinds": centers_config_params[center]["id"] + "_wind",
        "rmw": centers_config_params[center]["id"] + "_rmw"
        if center in ["USA", "REUNION", "BOM"]
        else None,
        "dist2land": "dist2land",
    }


def check_and_plot_track_data(track_data: xr.Dataset) -> plt.Figure:
    """
    Check the track data for missing values and plot the track.

    Parameters
    ----------
    track_data : xr.Dataset
        The track data to check in IBTrACS format. Must contain variables
        for time, longitude, latitude, pressure, and maximum winds for
        each center.

    Returns
    -------
    plt.Figure
        Figure object containing four subplots:
        1. Storm coordinates (lon vs lat)
        2. Minimum central pressure time series
        3. Maximum sustained winds time series
        4. Radii of maximum winds time series (if available)

    Notes
    -----
    - Longitude values are converted to [0-360º] convention
    - Wind speeds are converted to 1-minute average using center-specific factors
    - Variables that are entirely NaN are omitted from plots
    """

    fig, axes = plt.subplots(1, 4, figsize=(20, 4))

    for center in all_centers:
        # dictionary for IBTrACS center
        center_info = get_center_information(center=center)
        # get var time
        ytime = track_data[center_info["time"]].values
        # remove NaTs (time)
        st_lon = track_data[center_info["longitude"]].values[~np.isnat(ytime)]
        st_lat = track_data[center_info["latitude"]].values[~np.isnat(ytime)]
        st_prs = track_data[center_info["pressure"]].values[~np.isnat(ytime)]
        st_win = track_data[center_info["maxwinds"]].values[~np.isnat(ytime)]
        if center_info["rmw"]:
            st_rmw = track_data[center_info["rmw"]].values[~np.isnat(ytime)]
        st_tim = ytime[~np.isnat(ytime)]
        # longitude convention [0-360º]
        st_lon[st_lon < 0] += 360
        # winds are converted to 1-min avg
        st_win = st_win / centers_config_params[center]["windfac"]
        # plot center data (full NaN variables ommitted)
        color = centers_config_params[center]["color"]

        # Plot storm data
        axes[0].plot(st_lon, st_lat, "-", c=color, label=center)
        axes[0].set_title("Storm coordinates", fontweight="bold")
        axes[0].set_xlabel("Longitude (º)")
        axes[0].set_ylabel("Latitude (º)")
        if not np.isnan(st_prs).all():
            axes[1].plot(st_tim, st_prs, ".-", c=color, label=center)
            axes[1].set_title("Minimum central pressure [mbar]", fontweight="bold")
        if not np.isnan(st_win).all():
            axes[2].plot(st_tim, st_win, ".-", c=color, label=center)
            axes[2].set_title("Maximum sustained winds [kt]", fontweight="bold")
        if center_info["rmw"] and not np.isnan(st_rmw).all():
            axes[3].plot(st_tim, st_rmw, ".-", c=color, label=center)
            axes[3].set_title("Radii of max winds [nmile]", fontweight="bold")

    # plot attributes
    for i, ax in enumerate(axes):
        ax.legend(loc="upper left")
        if i > 0:
            ax.set_xticklabels(ax.get_xticklabels(), rotation=45)

    return fig


def filter_track_by_basin(tracks_data: xr.Dataset, id_basin: str) -> xr.Dataset:
    """
    Filter the tracks data by basin.

    Parameters
    ----------
    tracks_data : xr.Dataset
        The tracks data to filter in IBTrACS format.
        Must contain a 'basin' variable.
    id_basin : str
        The basin ID to filter by (e.g., 'NA', 'SP', etc.)

    Returns
    -------
    xr.Dataset
        The filtered tracks data containing only storms from the specified basin

    Examples
    --------
    >>> sp_tracks = filter_track_by_basin(tracks_data, "SP")
    >>> print(sp_tracks.storm.size)
    """

    # TODO: check whether just [0, :] is needed, as it is the only one used
    return tracks_data.isel(
        storm=np.where(tracks_data.basin.values[0, :].astype(str) == id_basin)
    )


def ibtracs_fit_pmin_wmax(ibtracs_data: xr.Dataset = None, N: int = 3) -> xr.Dataset:
    """
    Generate polynomial fit coefficients for pressure-wind relationship.

    Parameters
    ----------
    ibtracs_data : xr.Dataset
        The IBTrACS dataset containing storm track data. Must include pressure
        and wind variables for each center.
    N : int, optional
        The order of the polynomial fit. Default is 3.

    Returns
    -------
    xr.Dataset
        A dataset containing:
        - coef_fit : Polynomial coefficients (center, basin, N+1)
        - pres_data : Pressure data used for fitting
        - wind_data : Wind data used for fitting

    Notes
    -----
    - Maximum wind speeds are converted to 1-min average for all RSMC centers
    - The polynomial fit is of the form: p = a₀ + a₁w + a₂w² + ... + aₙwⁿ
      where p is pressure and w is wind speed

    Examples
    --------
    >>> fit_data = ibtracs_fit_pmin_wmax(ibtracs_data, N=3)
    >>> print(fit_data.coef_fit.shape)  # (n_centers, n_basins, N+1)
    """

    try:
        return xr.open_dataset(PATHS["SHYTCWAVES_COEFS"])

    except Exception as e:
        print(f"File could not be opened: {PATHS['SHYTCWAVES_COEFS']}. Error: {e}")

        coef_fit = np.nan * np.zeros((len(all_centers), len(all_basins), N + 1))
        pres_data = np.nan * np.zeros((len(all_centers), len(all_basins), 200000))
        wind_data = np.nan * np.zeros((len(all_centers), len(all_basins), 200000))

        for ic, center in enumerate(all_centers):
            center_info = get_center_information(center=center)

            for basin in center_info["basin"]:
                # filter tracks data by basin
                filtered_tracks = filter_track_by_basin(
                    tracks_data=ibtracs_data, id_basin=basin
                )
                # extract and reshape basin data: pressure, wind, landbasin
                newshape = filtered_tracks.storm.size * filtered_tracks.date_time.size
                landbasin = filtered_tracks[center_info["dist2land"]].values.reshape(
                    newshape
                )
                Pbasin = filtered_tracks[center_info["pressure"]].values.reshape(
                    newshape
                )
                Wbasin = filtered_tracks[center_info["maxwinds"]].values.reshape(
                    newshape
                )
                # winds are converted to 1-min avg [kt]
                Wbasin /= center_info["windfac"]
                PWbasin_s = np.column_stack((Pbasin, Wbasin, landbasin))
                # index for removing NaNs (including landmask)
                ix_nonan = ~np.isnan(PWbasin_s).any(axis=1)
                PWbasin_s = PWbasin_s[ix_nonan]
                # Fitting Polynomial Regression to the dataset
                X, y = PWbasin_s[:, 0], PWbasin_s[:, 1]
                u = polyfit(X, y, deg=N)
                # store coefficients
                ibasin = np.where(basin == np.array(all_basins))[0][0]
                coef_fit[ic, ibasin, :] = u
                pres_data[ic, ibasin, : PWbasin_s.shape[0]] = PWbasin_s[:, 0].T
                wind_data[ic, ibasin, : PWbasin_s.shape[0]] = PWbasin_s[:, 1].T

        # dataset
        xds = xr.Dataset(
            {
                "coef": (("center", "basin", "polynomial"), coef_fit),
                "pres": (("center", "basin", "data"), pres_data),
                "wind": (("center", "basin", "data"), wind_data),
            },
            {
                "center": np.asarray(all_centers),
                "basin": np.asarray(all_basins),
            },
        )
        xds.coef.attrs["name"] = "Fitting polynomial coefficients (Pmin, Wmax)"
        xds.coef.attrs["units"] = "Pressure (mbar), Wind speed (kt, 1-min avg)"

        return xds


def get_category(ycpres: np.ndarray) -> np.ndarray:
    """
    Calculate storm category based on central pressure.

    Parameters
    ----------
    ycpres : np.ndarray
        Array of central pressures in millibars (mbar).

    Returns
    -------
    np.ndarray
        Array of storm categories:
        - 6: Missing data (pressure = 0 or NaN)
        - 5: Category 5 (pressure < 920 mbar)
        - 4: Category 4 (920 ≤ pressure < 944 mbar)
        - 3: Category 3 (944 ≤ pressure < 964 mbar)
        - 2: Category 2 (964 ≤ pressure < 979 mbar)
        - 1: Category 1 (979 ≤ pressure < 1000 mbar)
        - 0: Tropical Storm/Depression (pressure ≥ 1000 mbar)

    Notes
    -----
    Uses np.select for vectorized operations. Categories are based on the
    Saffir-Simpson Hurricane Wind Scale pressure thresholds.

    Examples
    --------
    >>> pressures = np.array([915, 950, 980, 1005, np.nan])
    >>> get_category(pressures)
    array([5, 3, 1, 0, 6])
    """

    conditions = [
        (ycpres == 0) | (np.isnan(ycpres)),
        (ycpres < 920),
        (ycpres < 944),
        (ycpres < 964),
        (ycpres < 979),
        (ycpres < 1000),
        (ycpres >= 1000),
    ]
    choices = [6, 5, 4, 3, 2, 1, 0]

    return np.select(conditions, choices)


def wind2rmw(wmax: np.ndarray, vmean: np.ndarray, latitude: np.ndarray) -> np.ndarray:
    """
    Calculate radius of maximum winds using Knaff et al. (2016) formula.

    Parameters
    ----------
    wmax : np.ndarray
        Maximum sustained winds in knots
    vmean : np.ndarray
        Mean translational speed in knots
    latitude : np.ndarray
        Latitude of the storm center in degrees

    Returns
    -------
    np.ndarray
        Radius of maximum winds (RMW) in nautical miles

    Notes
    -----
    Implements the Knaff et al. (2016) formula for estimating RMW:
    1. Subtracts translational speed from observed maximum wind speed
    2. Converts 10m wind speed to gradient level wind speed using beta factor
    3. Applies empirical formula accounting for wind speed and latitude

    References
    ----------
    Knaff, J. A., et al. (2016). "Estimation of Tropical Cyclone Wind Structure
    Parameters from Satellite Imagery", Weather and Forecasting.

    Examples
    --------
    >>> rmw = wind2rmw(np.array([100]), np.array([10]), np.array([25]))
    array([26.51])
    """

    pifac = np.arccos(-1) / 180  # pi/180
    beta = 0.9  # conversion factor of wind speed

    # Subtract translational speed from observed maximum wind speed
    vkt = wmax - vmean  # [kt]

    # Convert 10m wind speed to gradient level wind speed
    vgrad = vkt / beta  # [kt]

    # Knaff et al. (2016) formula
    rm = (
        218.3784
        - 1.2014 * vgrad
        + np.power(vgrad / 10.9844, 2)
        - np.power(vgrad / 35.3052, 3)
        - 145.509 * np.cos(latitude * pifac)
    )  # [nmile]

    return rm


def get_vmean(
    lat1: Union[float, np.ndarray],
    lon1: Union[float, np.ndarray],
    lat2: Union[float, np.ndarray],
    lon2: Union[float, np.ndarray],
    deltat: Union[float, np.ndarray],
) -> Tuple[
    Union[float, np.ndarray],
    Union[float, np.ndarray],
    Union[float, np.ndarray],
    Union[float, np.ndarray],
]:
    """
    Calculate storm translation speed and direction between two points.

    Parameters
    ----------
    lat1 : Union[float, np.ndarray]
        Latitude of starting point(s) in degrees
    lon1 : Union[float, np.ndarray]
        Longitude of starting point(s) in degrees
    lat2 : Union[float, np.ndarray]
        Latitude of ending point(s) in degrees
    lon2 : Union[float, np.ndarray]
        Longitude of ending point(s) in degrees
    deltat : Union[float, np.ndarray]
        Time step(s) in hours

    Returns
    -------
    Tuple[Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]]
        Tuple containing:
        - gamma_h : Forward direction in degrees from North
        - vmean : Translation speed in km/h
        - vu : x-component of translation velocity in km/h
        - vy : y-component of translation velocity in km/h

    Notes
    -----
    Uses great circle distance calculation to determine distance between points.
    Translation speed is calculated by dividing distance by time step.
    Direction components are calculated using trigonometry.

    Examples
    --------
    >>> gamma, v, vx, vy = get_vmean(0, 0, 0, 1, 6)
    (270.0, 18.55, 18.55, 5.68e-15)
    """

    arcl_h, gamma_h = geodesic_distance_azimuth(lat2, lon2, lat1, lon1)  # great circle

    r = arcl_h * np.pi / 180.0 * EARTH_RADIUS  # distance between coordinates [km]
    vmean = r / deltat  # translation speed [km/h]

    vx = vmean * np.sin((gamma_h + 180) * np.pi / 180)
    vy = vmean * np.cos((gamma_h + 180) * np.pi / 180)

    return gamma_h, vmean, vx, vy  # [º], [km/h]


def wind2pres(
    xds_coef: xr.Dataset, st_wind: np.ndarray, st_center: str, st_basin: str
) -> np.ndarray:
    """
    Convert maximum wind speeds to minimum central pressure using fitted coefficients.
    As many other functions in this module, this works for IBTrACS data for the moment.

    Parameters
    ----------
    xds_coef : xr.Dataset
        Dataset containing polynomial fitting coefficients for Pmin-Wmax relationship.
    st_wind : np.ndarray
        Storm maximum winds in knots (1-min average).
    st_center : str
        Storm center/agency identifier (e.g., 'WMO', 'TOKYO').
    st_basin : str
        Storm basin identifier (e.g., 'NA', 'WP').

    Returns
    -------
    np.ndarray
        Predicted minimum central pressure values in millibars

    Notes
    -----
    Uses polynomial regression coefficients to estimate minimum central pressure
    from maximum wind speeds. The polynomial is of the form:
    p = a₀ + a₁w + a₂w² + ... + aₙwⁿ
    where p is pressure and w is wind speed.

    Examples
    --------
    >>> pres = wind2pres(coef_dataset, np.array([100]), 'WMO', 'NA')
    array([955.2])
    """

    pmin_pred = []

    for iwind in st_wind:
        # select Pmin-Wmax coefficients
        coeff = xds_coef.sel(center=st_center, basin=st_basin).coef.values[:]

        # aisle equation
        coeff[-1] -= iwind

        # pressure root
        pmin_pred.append(
            np.real(np.roots(coeff))[np.where(np.imag(np.roots(coeff)) == 0)[0][0]]
        )

    return np.array(pmin_pred)  # [mbar]


def resample_storm_6h(storm: xr.Dataset) -> xr.Dataset:
    """
    Resample storm data to 6-hour intervals.

    Parameters
    ----------
    storm : xr.Dataset
        Storm dataset containing time series data with arbitrary time intervals

    Returns
    -------
    xr.Dataset
        Resampled storm dataset with 6-hour intervals

    Notes
    -----
    This function:
    1. Removes NaT (Not a Time) values from time coordinate
    2. Removes NaN values from WMO pressure data
    3. Resets lon/lat coordinates
    4. Resamples all variables to 6-hour intervals using linear interpolation
    5. Preserves the original basin identifier

    Examples
    --------
    >>> resampled = resample_storm_6h(storm_data)
    >>> np.all(np.diff(resampled.time) == np.timedelta64(6, 'h'))
    True
    """

    storm = storm.isel(date_time=~np.isnat(storm.time.values))
    storm = storm.isel(date_time=~np.isnan(storm.wmo_pres.values))
    storm = storm.reset_coords(["lon", "lat"])
    ibasin = storm.basin.values[0]
    storm = (
        storm.swap_dims({"date_time": "time"}).resample(time="6H").interpolate("linear")
    )
    storm["basin"] = (("time"), np.array([ibasin] * storm.time.size))

    return storm


def historic_track_preprocessing(
    xds: xr.Dataset,
    center: str = "WMO",
    forecast_on: bool = False,
    database_on: bool = False,
    st_param: bool = False,
) -> pd.DataFrame:
    """
    Preprocess historical storm track data from IBTrACS or forecast sources.

    Parameters
    ----------
    xds : xr.Dataset
        Historical storm track dataset with storm dimension.
    center : str, optional
        IBTrACS center code (e.g., 'WMO', 'TOKYO'). Default is "WMO".
    forecast_on : bool, optional
        Whether track is forecasted (not IBTrACS). Default is False.
    database_on : bool, optional
        Whether to keep data only at 0,6,12,18 hours. Default is False.
    st_param : bool, optional
        Whether to keep data as original. Default is False.

    Returns
    -------
    pd.DataFrame
        Preprocessed storm track data with columns:
        - center : Storm center identifier
        - basin : Storm basin identifier
        - dist2land : Distance to nearest land (km)
        - longitude : Storm longitude (0-360°)
        - latitude : Storm latitude
        - move : Forward direction (degrees)
        - mean_velocity : Translation speed (kt)
        - pressure : Central pressure (mbar)
        - maxwinds : Maximum winds (kt, 1-min avg)
        - rmw : Radius of maximum winds (nmile)
        - category : Storm category (0-6)
        - timestep : Time step (hours)
        - storm_vmean : Mean translation speed (kt)

    Notes
    -----
    Processing steps include:
    1. Removing NaT/NaN values from time and pressure data
    2. Converting longitudes to [0°-360°] convention
    3. Rounding dates to nearest hour
    4. Calculating translation speed and direction
    5. Converting wind speeds to 1-min average
    6. Determining storm category based on pressure

    For forecast tracks:
    - Basin IDs are converted to IBTrACS equivalents
    - Missing pressures are estimated from wind speeds using fitted coefficients
    - Southern hemisphere basins are determined by latitude and longitude

    Examples
    --------
    >>> df = historic_track_preprocessing(track_data, center='WMO')
    >>> print(f"Track duration: {len(df)} time steps")
    Track duration: 48
    >>> print(f"Storm category: {df['category'].iloc[0]}")
    Storm category: 3
    """

    # dictionary for IBTrACS center
    d_vns = get_center_information(center=center)

    # get names of variables
    nm_tim = d_vns["time"]
    nm_bas = d_vns["basin"]
    nm_lon = d_vns["longitude"]
    nm_lat = d_vns["latitude"]
    nm_prs = d_vns["pressure"]
    nm_win = d_vns["maxwinds"]
    nm_rmw = d_vns["rmw"]

    # get var time
    ytime = xds[nm_tim].values  # dates format: datetime64

    # remove NaTs (time)
    ydist = xds["dist2land"].values[~np.isnat(ytime)]  # distance to land [km]
    ybasin = xds[nm_bas].values[~np.isnat(ytime)]  # basin
    ylat_tc = xds[nm_lat].values[~np.isnat(ytime)]  # latitude
    ylon_tc = xds[nm_lon].values[~np.isnat(ytime)]  # longitude
    ycpres = xds[nm_prs].values[~np.isnat(ytime)]  # pressure [mbar]
    ywind = xds[nm_win].values[~np.isnat(ytime)]  # wind speed [kt]
    if nm_rmw:
        yradii = xds[nm_rmw].values[~np.isnat(ytime)]  # rmw [nmile]
    ytime = ytime[~np.isnat(ytime)]

    # remove common NaNs (pressure & wind)
    posnonan_p_w = np.unique(
        np.concatenate((np.argwhere(~np.isnan(ycpres)), np.argwhere(~np.isnan(ywind))))
    )
    ytime = ytime[posnonan_p_w]
    ydist = ydist[posnonan_p_w]
    ybasin = ybasin[posnonan_p_w]
    ylat_tc = ylat_tc[posnonan_p_w]
    ylon_tc = ylon_tc[posnonan_p_w]
    ycpres = ycpres[posnonan_p_w]
    ywind = ywind[posnonan_p_w]
    if nm_rmw:
        yradii = yradii[posnonan_p_w]
    if not nm_rmw:
        yradii = np.full(ycpres.size, np.nan)  # [np.nan] * ycpres.size

    ###########################################################################
    # forecast input may lack pressure
    if forecast_on:
        # convert basin ids (IBTrACS equivalent)
        dict_basin_forecast = {
            "AL": "NA",
            "LS": "SA",
            "EP": "EP",
            "CP": "WP",
            "WP": "WP",
            "SH": "SP",
            "IO": "NI",
        }
        ybasin = np.array([dict_basin_forecast[ybas] for ybas in ybasin])
        ybasin[(ylat_tc < 0) & (ylon_tc < 135)] = "SI"

        if np.isnan(ycpres).any():
            # ibtracs fitting coefficients path
            xds_coef = ibtracs_fit_pmin_wmax()

            # fill pressure gaps
            ycpres[np.isnan(ycpres)] = wind2pres(
                xds_coef, ywind[np.isnan(ycpres)], center, ybasin[0]
            )

        ybasin = np.array([np.bytes_(ybas) for ybas in ybasin])
    ###########################################################################

    # remove NaNs (pressure)
    ytime = ytime[~np.isnan(ycpres)]
    st_dist2land = ydist[~np.isnan(ycpres)]
    ybasin = ybasin[~np.isnan(ycpres)]
    st_lat = ylat_tc[~np.isnan(ycpres)]
    st_lon = ylon_tc[~np.isnan(ycpres)]
    st_pres = ycpres[~np.isnan(ycpres)]
    st_wind = ywind[~np.isnan(ycpres)]
    if nm_rmw:
        st_rmw = yradii[~np.isnan(ycpres)]
    if not nm_rmw:
        st_rmw = np.full(st_pres.size, np.nan)  # [np.nan] * st_pres.size

    # conflicting times
    dft = pd.DataFrame({"time": ytime})
    dft["time"] = pd.to_datetime(dft["time"], format="%Y-%m-%d %H:%M:%S").dt.round("1h")
    hr = dft["time"].dt.hour.values

    # duplicate times
    pos_out = np.where(np.diff(hr) == 0)[0]

    # keep 0,3,6,9,12,15,18 hours (remove in between data)
    # TODO: option to interpolate to fixed target hours
    if database_on:
        pos_in = np.where((hr == 0) | (hr == 6) | (hr == 12) | (hr == 18))[0]
    else:
        pos_in = np.where(
            (hr == 0)
            | (hr == 6)
            | (hr == 12)
            | (hr == 18)
            | (hr == 3)
            | (hr == 9)
            | (hr == 15)
            | (hr == 21)
        )[0]
    # for parameterized tracks hours can be random
    if st_param:
        pos_in = np.where(hr > 0)

    # remove duplicates
    pos = np.setdiff1d(pos_in, pos_out)  # pos_in[pos_in != pos_out].ravel()
    ytime = ytime[pos]
    st_dist2land = st_dist2land[pos]
    ybasin = ybasin[pos]
    st_lat = st_lat[pos]
    st_lon = st_lon[pos]
    st_pres = st_pres[pos]
    st_wind = st_wind[pos]
    st_rmw = st_rmw[pos]

    # only when storm data available
    if st_pres.size > 0:
        # longitude convention: [0º,360º]
        st_lon[st_lon < 0] = st_lon[st_lon < 0] + 360

        # round dates to hour ---> half hour
        round_to = 3600 / 2
        st_time = []
        for i in range(len(ytime)):
            dt = ytime[i].astype("datetime64[s]").tolist()
            seconds = (dt - dt.min).seconds
            rounding = (seconds + round_to / 2) // round_to * round_to
            out = dt + timedelta(0, rounding - seconds, -dt.microsecond)
            st_time.append(out)
        st_time = np.asarray(st_time)

        # storm coordinates timestep [hours]
        ts = st_time[1:] - st_time[:-1]
        ts = [ts[i].total_seconds() / 3600 for i in range(ts.size)]
        ts.append(np.nan)

        # calculate Vmean
        st_vmean, st_move = [], []
        for i in range(0, len(st_time) - 1):
            # consecutive storm coordinates
            lon1, lon2 = st_lon[i], st_lon[i + 1]
            lat1, lat2 = st_lat[i], st_lat[i + 1]

            # translation speed
            gamma_h, vel_mean, _, _ = get_vmean(lat1, lon1, lat2, lon2, ts[i])
            st_vmean.append(vel_mean / 1.852)  # translation speed [km/h to kt]
            st_move.append(gamma_h)  # forward direction [º]
        st_vmean.append(np.nan)
        st_move.append(np.nan)

        # mean value
        vmean = np.nanmean(st_vmean)  # [kt]

        # storm category
        categ = get_category(st_pres)

        # convert (X)-min to 1-min avg winds (depends on center)
        st_wind = st_wind / centers_config_params[center]["windfac"]

        # get basin string
        st_basin = [str(c.decode("UTF-8")) for c in ybasin]

        # store storm variables
        df = pd.DataFrame(
            index=st_time,
            columns=[
                "center",
                "basin",
                "dist2land",
                "longitude",
                "latitude",
                "move",
                "mean_velocity",
                "pressure",
                "maxwinds",
                "rmw",
                "category",
                "timestep",
                "storm_vmean",
            ],
        )

        df["center"] = center
        df["basin"] = st_basin
        if "dist2land" in xds.keys():
            df["dist2land"] = st_dist2land
        else:
            df["dist2land"] = np.nan
        df["longitude"] = st_lon
        df["latitude"] = st_lat
        df["move"] = st_move
        df["mean_velocity"] = st_vmean
        df["pressure"] = st_pres
        df["maxwinds"] = st_wind
        df["rmw"] = st_rmw
        df["category"] = categ
        df["timestep"] = ts
        df["storm_vmean"] = vmean

        df.attrs = {
            "dist2land": "km",
            "velocity": "kt",
            "pressure": "mbar",
            "maxwinds": "kt, 1-min average sustained winds (converted from X-min)",
            "rmw": "nmile",
            "timestep": "hour",
        }

    else:
        df = pd.DataFrame(
            index=[],
            columns=[
                "center",
                "basin",
                "dist2land",
                "longitude",
                "latitude",
                "move",
                "mean_velocity",
                "pressure",
                "maxwinds",
                "rmw",
                "category",
                "timestep",
                "storm_vmean",
            ],
        )

    return df


def historic_track_interpolation(
    df: pd.DataFrame,
    dt_comp: float,
    y0: float = None,
    x0: float = None,
    great_circle: bool = True,
    wind_estimate_on: bool = False,
    fit: bool = False,
    interpolation: bool = True,
    mode: str = "first",
    radi_estimate_on: bool = True,
) -> Tuple[pd.DataFrame, np.ndarray]:
    """
    Interpolate storm track variables to computational time steps.

    Parameters
    ----------
    df : pd.DataFrame
        Storm track DataFrame with historical data.
    dt_comp : float
        Computation time step in minutes.
    y0 : float, optional
        Target latitude coordinate. Default is None.
    x0 : float, optional
        Target longitude coordinate. Default is None.
    great_circle : bool, optional
        Whether to use great circle distances. Default is True.
    wind_estimate_on : bool, optional
        Whether to use empirical estimates instead of historical winds. Default is False.
    fit : bool, optional
        Whether to estimate winds when wind=0. Default is False.
    interpolation : bool, optional
        Whether to interpolate storm variables. Default is True.
    mode : str, optional
        Value selection for constant segments ('first' or 'mean'). Default is "first".
    radi_estimate_on : bool, optional
        Whether to estimate missing RMW values. Default is True.

    Returns
    -------
    Tuple[pd.DataFrame, np.ndarray]
        - DataFrame with interpolated storm track variables.
        - Array of interpolated time coordinates.

    Notes
    -----
    The function:
    1. Interpolates track points to match computational time step
    2. Estimates missing values using empirical relationships:
        - Maximum winds from pressure (Pmin-Wmax relationship)
        - Radius of maximum winds (Knaff et al. 2016)
    3. Can use either linear interpolation or constant segments
    4. Preserves metadata and adds computational parameters

    Examples
    --------
    >>> st, times = historic_track_interpolation(track_data, dt_comp=30)
    >>> print(f"Interpolated points: {len(st)}")
    Interpolated points: 144
    """

    # historic storm variables
    st_time = df.index.values[:]  # datetime format
    st_center = df["center"].values[0]
    st_basin = df["basin"].values[:]
    st_dist2land = df["dist2land"].values[:]  # [km]
    st_lon = df["longitude"].values[:]
    st_lat = df["latitude"].values[:]
    st_vmean = df["mean_velocity"].values[:]  # [kt]
    st_pres = df["pressure"].values[:]  # [mbar]
    st_wind = df["maxwinds"].values[:]  # [kt, 1-min avg]
    st_rmw = df["rmw"].values[:]  # [nmile]
    st_step = df["timestep"].values[:]  # [hour]
    wind_fill = np.full(st_pres.size, False)
    rmw_fill = np.full(st_pres.size, False)

    # select Pmin-Wmax polynomial fitting coefficients (IBTrACS center,basin)
    xds_coef = ibtracs_fit_pmin_wmax()
    try:
        coefs = xds_coef.sel(
            center=st_center,
            basin=st_basin,
        ).coef.values[:]
    except Exception:
        coefs = xds_coef.sel(
            center=st_center.encode("utf-8"),
            basin=st_basin.astype("bytes"),
        ).coef.values[:]

    p1, p2, p3, p4 = coefs[:, 0], coefs[:, 1], coefs[:, 2], coefs[:, 3]
    wind_estimate = (
        p1 * np.power(st_pres, 3)
        + p2 * np.power(st_pres, 2)
        + p3 * np.power(st_pres, 1)
        + p4
    )

    # maxwinds gaps filled with Pmin-Wmax coefficients
    posnan = np.argwhere(np.isnan(st_wind))  # np.where(st_wind==np.nan)[0]
    poszero = np.where(st_wind == 0)[0]
    if np.isnan(st_wind).all() or wind_estimate_on:  # wind NOT provided
        wind_fill = np.full(st_wind.size, True)
        st_wind = wind_estimate
    ###########################################################################
    elif np.isnan(st_wind).any():  # wind with some NaNs
        if fit and posnan.size > 0:  # data filled (for wind=0)
            wind_fill[posnan] = True
            st_wind[posnan] = wind_estimate[posnan]
    elif fit and poszero.size > 0:  # wind provided
        wind_fill[poszero] = True
        st_wind[poszero] = wind_estimate[poszero]
    ###########################################################################
    #    else:                                             # wind provided
    #        pos = np.where(st_wind==0)[0]
    #        if fit and pos.size>0:    # data filled (for wind=0)
    #            wind_fill[pos] = True
    #            st_wind[pos]   = wind_estimate[pos]

    # radii of maximum winds gaps filled with Knaff et al. (2016) estimate
    rmw_estimate = wind2rmw(st_wind, st_vmean, st_lat)
    if radi_estimate_on:
        if np.isnan(st_rmw).all():  # all
            rmw_fill = np.full(rmw_estimate.size, True)
            st_rmw = rmw_estimate
        elif np.isnan(st_rmw).any():  # some
            rmw_fill[np.isnan(st_rmw)] = True
            st_rmw[np.isnan(st_rmw)] = rmw_estimate[np.isnan(st_rmw)]
    else:
        if np.isnan(st_rmw).any():
            st_rmw = df.rmw.interpolate(method="time").values

    # storm variables (original values or mean values)
    if mode == "mean":
        st_pres = np.mean((st_pres, np.append(st_pres[1:], np.nan)), axis=0)
        st_wind = np.mean((st_wind, np.append(st_wind[1:], np.nan)), axis=0)
        st_rmw = np.mean((st_rmw, np.append(st_rmw[1:], np.nan)), axis=0)

    # number of time steps between consecutive interpolated storm coordinates
    # to match SWAN computational timestep
    ts_h = st_step  # hours
    nts = np.asarray(st_step) * 60 / dt_comp  # number of intervals

    # initialize interpolated variables
    bas, dist, lon, lat, move, vmean, vu, vy = [], [], [], [], [], [], [], []
    p0, vmax, rmw, vmaxfill, rmwfill = [], [], [], [], []
    time_input = np.empty((0,), dtype="datetime64[ns]")

    for i in range(0, len(st_time) - 1):
        # time array for SWAN input
        date_ini = st_time[i]
        time_input0 = pd.date_range(
            date_ini, periods=int(nts[i]), freq="{0}min".format(dt_comp)
        )
        time_input = np.append(np.array(time_input), np.array(time_input0))

        # consecutive storm coordinates
        lon1, lon2 = st_lon[i], st_lon[i + 1]
        lat1, lat2 = st_lat[i], st_lat[i + 1]

        # translation speed
        arcl_h, gamma_h = geodesic_distance_azimuth(lat2, lon2, lat1, lon1)
        r = arcl_h * np.pi / 180.0 * EARTH_RADIUS  # distance between coordinates [km]
        dx = r / nts[i]  # distance during time step [km]
        tx = ts_h[i] / nts[i]  # time period during time step [h]
        vx = float(dx) / tx / 3.6  # translation speed [km/h to m/s]
        vx = vx / 0.5144  # translation speed [m/s to kt]

        # interpolate at timesteps
        for j in range(int(nts[i])):
            bas.append(st_basin[i])
            dist.append(st_dist2land[i])

            # append interpolated lon, lat
            if not great_circle:
                glon = (
                    lon1
                    - (dx * 180 / (EARTH_RADIUS * np.pi))
                    * np.sin(gamma_h * np.pi / 180)
                    * j
                )
                glat = (
                    lat1
                    - (dx * 180 / (EARTH_RADIUS * np.pi))
                    * np.cos(gamma_h * np.pi / 180)
                    * j
                )
            else:
                glon, glat, baz = shoot(lon1, lat1, gamma_h + 180, float(dx) * j)
            lon.append(glon)
            lat.append(glat)

            # append storm track parameters
            move.append(gamma_h)
            vmean.append(vx)
            vu.append(vx * np.sin((gamma_h + 180) * np.pi / 180))
            vy.append(vx * np.cos((gamma_h + 180) * np.pi / 180))

            # append pmin, wind, rmw (interpolated/constant)
            if interpolation:
                p0.append(st_pres[i] + j * (st_pres[i + 1] - st_pres[i]) / nts[i])
                vmax.append(st_wind[i] + j * (st_wind[i + 1] - st_wind[i]) / nts[i])
                rmw.append(st_rmw[i] + j * (st_rmw[i + 1] - st_rmw[i]) / nts[i])
                vmaxfill.append(wind_fill[i] or wind_fill[i + 1])
                rmwfill.append(rmw_fill[i] or rmw_fill[i + 1])
            else:
                p0.append(st_pres[i])
                vmax.append(st_wind[i])
                rmw.append(st_rmw[i])
                vmaxfill.append(wind_fill[i])
                rmwfill.append(rmw_fill[i])

    # longitude convention [0º,360º]
    lon = np.array(lon)
    lon[lon < 0] = lon[lon < 0] + 360

    # method attribute
    n_hour = dt_comp / 60
    if interpolation:
        method = "interpolation ({0}h)".format(n_hour)
    else:
        if mode == "mean":
            method = "segment ({0}h, mean value)".format(n_hour)
        elif mode == "first":
            method = "segment ({0}h, node value)".format(n_hour)

    # storm track (pd.DataFrame)
    st = pd.DataFrame(
        index=time_input,
        columns=[
            "center",
            "basin",
            "dist2land",
            "lon",
            "lat",
            "move",
            "vf",
            "vfx",
            "vfy",
            "pn",
            "p0",
            "vmax",
            "rmw",
            "vmaxfill",
            "rmwfill",
        ],
    )

    st["center"] = st_center
    st["basin"] = bas
    st["dist2land"] = dist  # distance to nearest land [km]
    st["lon"] = lon  # longitude coordinate
    st["lat"] = lat  # latitude coordinate
    st["move"] = move  # gamma, forward direction
    st["vf"] = vmean  # translational speed [kt]
    st["vfx"] = vu  # x-component
    st["vfy"] = vy  # y-component
    st["pn"] = 1013  # average pressure at the surface [mbar]
    st["p0"] = p0  # minimum central pressure [mbar]
    st["vmax"] = vmax  # maximum winds [kt, 1-min avg]
    st["rmw"] = rmw  # radii of maximum winds [nmile]
    st["vmaxfill"] = vmaxfill  # True if maxwinds filled with Pmin-Vmax
    st["rmwfill"] = rmwfill  # True if radii filled with Knaff 2016

    # add metadata
    st.attrs = {
        "method": method,
        "center": st_center,
        "dist": "km",
        "vf": "kt",
        "p0": "mbar",
        "vmax": "kt, 1-min avg",
        "rmw": "nmile",
        "x0": x0,
        "y0": y0,
        "R": 4,
    }

    return st, time_input


def track_triming(
    st: pd.DataFrame, lat00: float, lon00: float, lat01: float, lon01: float
) -> pd.DataFrame:
    """
    Trim storm track to specified geographical bounds.

    Parameters
    ----------
    st : pd.DataFrame
        Storm track DataFrame containing time series data.
    lat00 : float
        Southern latitude bound in degrees.
    lon00 : float
        Western longitude bound in degrees.
    lat01 : float
        Northern latitude bound in degrees.
    lon01 : float
        Eastern longitude bound in degrees.

    Returns
    -------
    pd.DataFrame
        Trimmed storm track containing only points within the specified bounds
        and preserving continuous time segments.

    Notes
    -----
    The function:
    1. Identifies all track points within the specified bounds
    2. Finds the earliest and latest times for points within bounds
    3. Returns all track points between these times to maintain continuity
    4. Preserves original metadata in the returned DataFrame

    Examples
    --------
    >>> trimmed = track_triming(track_data, 20, -80, 30, -60)
    >>> print(f"Points in bounds: {len(trimmed)}")
    Points in bounds: 24
    """

    lo, la = st.lon.values[:], st.lat.values[:]

    # select track coordinates within the target domain area
    st_trim = st.iloc[(lo <= lon01) & (lo >= lon00) & (la <= lat01) & (la >= lat00)]

    # get min/max times within the area
    tmin = np.min(st_trim.index.values)
    tmax = np.max(st_trim.index.values)

    # select all the time window
    st_trim = st.iloc[(st.index >= tmin) & (st.index <= tmax)]

    # add metadata
    st_trim.attrs = st.attrs

    return st_trim


def track_triming_circle(
    st: pd.DataFrame, plon: float, plat: float, radii: float
) -> pd.DataFrame:
    """
    Trim storm track to points within a circular domain.

    Parameters
    ----------
    st : pd.DataFrame
        Storm track DataFrame containing time series data.
    plon : float
        Longitude of circle center in degrees.
    plat : float
        Latitude of circle center in degrees.
    radii : float
        Radius of circular domain in degrees.

    Returns
    -------
    pd.DataFrame
        Trimmed storm track containing only points within the circular domain
        and preserving continuous time segments.

    Notes
    -----
    The function:
    1. Calculates geodesic distance from each track point to circle center
    2. Identifies points within the specified radius
    3. Returns continuous time segment containing all points within radius
    4. Preserves original metadata in the returned DataFrame

    Examples
    --------
    >>> trimmed = track_triming_circle(track_data, -75, 25, 5)
    >>> print(f"Points in circle: {len(trimmed)}")
    Points in circle: 18
    """

    lo, la = st.lon.values[:], st.lat.values[:]

    # stack storm longitude, latitude
    lonlat_s = np.column_stack((lo, la))

    # calculate geodesic distance (degree)
    geo_dist = []
    for lon_ps, lat_ps in lonlat_s:
        geo_dist.append(geodesic_distance(lat_ps, lon_ps, plat, plon))
    geo_dist = np.asarray(geo_dist)

    # find storm inside circle and calculate parameters
    if (geo_dist < radii).any() & (geo_dist.size > 1):
        # storm inside circle
        ix_in = np.where(geo_dist < radii)[0][:]
        # select track coordinates within the target domain area
        st_trim = st.iloc[ix_in]

    else:
        st_trim = st.iloc[st.lat.values[:] > 90]

    # get min/max times within the area
    tmin = np.min(st_trim.index.values)
    tmax = np.max(st_trim.index.values)

    # select all the time window
    st_trim = st.iloc[(st.index >= tmin) & (st.index <= tmax)]

    # add metadata
    st_trim.attrs = st.attrs

    return st_trim


def stopmotion_trim_circle(
    df_seg: pd.DataFrame, plon: float, plat: float, radii: float
) -> pd.DataFrame:
    """
    Trim storm track segments to those intersecting a circular domain.

    Parameters
    ----------
    df_seg : pd.DataFrame
        DataFrame containing storm track segments with start/end coordinates.
    plon : float
        Longitude of circle center in degrees.
    plat : float
        Latitude of circle center in degrees.
    radii : float
        Radius of circular domain in degrees.

    Returns
    -------
    pd.DataFrame
        Trimmed DataFrame containing only segments that intersect the circular domain.

    Notes
    -----
    The function:
    1. Calculates geodesic distances from both endpoints of each segment to circle center
    2. Takes minimum distance to determine if segment intersects circle
    3. Returns continuous time segment containing all intersecting segments
    4. Preserves original metadata in the returned DataFrame

    Examples
    --------
    >>> trimmed = stopmotion_trim_circle(segments, -75, 25, 5)
    >>> print(f"Segments intersecting circle: {len(trimmed)}")
    Segments intersecting circle: 8
    """

    lo0, la0 = df_seg.lon_i.values[:], df_seg.lat_i.values[:]
    lo1, la1 = df_seg.lon_t.values[:], df_seg.lat_t.values[:]

    # stack storm longitude, latitude
    lonlat_s0 = np.column_stack((lo0, la0))
    lonlat_s1 = np.column_stack((lo1, la1))

    # calculate geodesic distance (degree)
    geo_dist0, geo_dist1 = [], []
    for lon_ps, lat_ps in lonlat_s0:
        geo_dist0.append(geodesic_distance(lat_ps, lon_ps, plat, plon))
    geo_dist0 = np.asarray(geo_dist0)
    for lon_ps, lat_ps in lonlat_s1:
        geo_dist1.append(geodesic_distance(lat_ps, lon_ps, plat, plon))
    geo_dist1 = np.asarray(geo_dist1)

    # get minimum distance of the two target endpoints
    geo_dist = np.min((geo_dist0, geo_dist1), axis=0)

    # find storm inside circle and calculate parameters
    if (geo_dist < radii).any() & (geo_dist.size > 1):
        # storm inside circle
        ix_in = np.where(geo_dist < radii)[0][:]
        # select track coordinates within the target domain area
        st_trim = df_seg.iloc[ix_in]

        # get min/max times within the area
        tmin = np.min(st_trim.index.values)
        tmax = np.max(st_trim.index.values)

        # select all the time window
        st_trim = df_seg.iloc[(df_seg.index >= tmin) & (df_seg.index <= tmax)]

        # add metadata
        st_trim.attrs = df_seg.attrs

    else:  # empty dataframe
        st_trim = pd.DataFrame(columns=df_seg.columns)

    return st_trim


def track_extent(
    st: pd.DataFrame, time_input: np.ndarray, dt_comp: float, time_extent: float = 48.0
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Extend storm track data for wave propagation analysis.

    Parameters
    ----------
    st : pd.DataFrame
        Storm track DataFrame containing time series data.
    time_input : np.ndarray
        Array of time coordinates.
    dt_comp : float
        Computational time step in minutes.
    time_extent : float, optional
        Additional time to extend simulation in hours. Default is 48.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        Tuple containing:
        - st_new : Extended storm track DataFrame.
        - we : Wave event DataFrame with empty fields for wave parameters.

    Notes
    -----
    The function:
    1. Calculates total simulation period (storm + propagation)
    2. Creates extended time array with specified time step
    3. Fills storm data in initial period
    4. Creates empty wave event DataFrame for entire period
    5. Preserves metadata and adds computational time step override

    Examples
    --------
    >>> st_ext, wave_evt = track_extent(track_data, times, 30, time_extent=24)
    >>> print(f"Extended duration: {len(st_ext)} time steps")
    Extended duration: 96 time steps
    """

    # total simulation period (hours): storm + propagation
    time_storm = (time_input[-1] - time_input[0]).astype(
        "timedelta64[h]"
    ) / np.timedelta64(1, "h")
    T = time_storm + time_extent

    # original track period
    size_ini = st.index.size
    date_ini = st.index[0]

    # total simulation period
    size_new = int(T * 60 / dt_comp)
    time_input_new = pd.date_range(
        date_ini, periods=size_new, freq="{0}MIN".format(dt_comp)
    )
    st_new = pd.DataFrame(index=time_input_new, columns=list(st.keys()))

    # combine
    st_new.values[:size_ini, :] = st.values
    if st.attrs["x0"]:
        st_new.x0 = st.attrs["x0"]
    if st.attrs["y0"]:
        st_new.y0 = st.attrs["y0"]

    # [OPTIONAL] override SWAN storm case computational delta time
    st_new.attrs["override_dtcomp"] = "{0}MIN".format(dt_comp)

    # generate wave event (empty)
    we = pd.DataFrame(
        index=time_input_new, columns=["hs", "t02", "dir", "spr", "U10", "V10"]
    )
    we["level"] = 0
    we["tide"] = 0

    return st_new, we


###############################################################################
# HYTCWAVES
# PARAMETERIZED storm tracks
# Given the storm parameters (HyTCWaves methodology) the track coordinates are
# calculated at each swan computational timestep
###############################################################################


def entrance_coords(
    delta: float,
    gamma: float,
    x0: float,
    y0: float,
    R: float,
    lon0: float,
    lon1: float,
    lat0: float,
    lat1: float,
) -> Tuple[float, float]:
    """
    Calculate storm entrance coordinates at computational domain boundary.

    Parameters
    ----------
    delta : float
        Storm track angle parameter in degrees.
    gamma : float
        Storm forward direction in degrees.
    x0 : float
        Site longitude coordinate in degrees.
    y0 : float
        Site latitude coordinate in degrees.
    R : float
        Radius of influence in degrees.
    lon0 : float
        Western longitude bound of computational domain.
    lon1 : float
        Eastern longitude bound of computational domain.
    lat0 : float
        Southern latitude bound of computational domain.
    lat1 : float
        Northern latitude bound of computational domain.

    Returns
    -------
    Tuple[float, float]
        Tuple containing:
        - x1 : Entrance longitude coordinate in degrees.
        - y1 : Entrance latitude coordinate in degrees.

    Notes
    -----
    The function:
    1. Calculates entrance point on the radius of influence
    2. Determines angles to domain corners from entrance point
    3. Uses storm direction to determine intersection with domain boundary
    4. Returns coordinates where storm track enters computational domain

    Examples
    --------
    >>> x1, y1 = entrance_coords(45, 270, -75, 25, 5, -80, -70, 20, 30)
    >>> print(f"Entrance point: ({x1:.1f}°, {y1:.1f}°)")
    Entrance point: (-78.5°, 27.5°)
    """

    # enter point in the radius
    xc = x0 + R * np.sin(delta * np.pi / 180)
    yc = y0 + R * np.cos(delta * np.pi / 180)

    # calculate angles that determine the storm boundary entrance  [degrees]
    ang_1 = np.arctan((lon1 - xc) / (lat1 - yc)) * 180 / np.pi  # upper right corner
    ang_2 = np.arctan((lon1 - xc) / (lat0 - yc)) * 180 / np.pi + 180  # lower right
    ang_3 = np.arctan((lon0 - xc) / (lat0 - yc)) * 180 / np.pi + 180  # lower left
    ang_4 = np.arctan((lon0 - xc) / (lat1 - yc)) * 180 / np.pi + 360  # upper left

    if (gamma > ang_1) & (gamma < ang_2):
        x1 = lon1
        d = (x1 - xc) / np.sin(gamma * np.pi / 180)
        y1 = yc + d * np.cos(gamma * np.pi / 180)

    elif (gamma > ang_2) & (gamma < ang_3):
        y1 = lat0
        d = (y1 - yc) / np.cos(gamma * np.pi / 180)
        x1 = xc + d * np.sin(gamma * np.pi / 180)

    elif (gamma > ang_3) & (gamma < ang_4):
        x1 = lon0
        d = (x1 - xc) / np.sin(gamma * np.pi / 180)
        y1 = yc + d * np.cos(gamma * np.pi / 180)

    elif (gamma > ang_4) | (gamma < ang_1):
        y1 = lat1
        d = (y1 - yc) / np.cos(gamma * np.pi / 180)
        x1 = xc + d * np.sin(gamma * np.pi / 180)

    return x1, y1


def track_site_parameters(
    step: float,
    pmin: float,
    vmean: float,
    delta: float,
    gamma: float,
    x0: float,
    y0: float,
    lon0: float,
    lon1: float,
    lat0: float,
    lat1: float,
    R: float,
    date_ini: str,
    center: str = "WMO",
    basin: str = "SP",
) -> pd.DataFrame:
    """
    Generate parameterized storm track within study area.

    Parameters
    ----------
    step : float
        Computational time step in minutes.
    pmin : float
        Minimum central pressure in millibars.
    vmean : float
        Mean translation speed in knots.
    delta : float
        Storm track angle parameter in degrees.
    gamma : float
        Storm forward direction in degrees.
    x0 : float
        Site longitude coordinate in degrees.
    y0 : float
        Site latitude coordinate in degrees.
    lon0 : float
        Western longitude bound of computational domain.
    lon1 : float
        Eastern longitude bound of computational domain.
    lat0 : float
        Southern latitude bound of computational domain.
    lat1 : float
        Northern latitude bound of computational domain.
    R : float
        Radius of influence in degrees.
    date_ini : str
        Initial date in format 'YYYY-MM-DD HH:MM'.
    center : str, optional
        IBTrACS center code. Default is "WMO".
    basin : str, optional
        Storm basin identifier. Default is "SP".

    Returns
    -------
    pd.DataFrame
        Storm track DataFrame with columns:
        - lon : Storm longitude coordinates
        - lat : Storm latitude coordinates
        - move : Forward direction (gamma)
        - vf : Translation speed (kt)
        - vfx : x-component of velocity
        - vfy : y-component of velocity
        - pn : Surface pressure (1013 mbar)
        - p0 : Minimum central pressure
        - vmax : Maximum winds (kt, 1-min avg)
        - rmw : Radius of maximum winds (nmile)

    Notes
    -----
    The function:
    1. Calculates entrance coordinates at domain boundary
    2. Generates track points at computational time steps
    3. Estimates maximum winds from pressure using fitted coefficients
    4. Estimates radius of maximum winds using Knaff et al. (2016)
    5. Preserves metadata including site coordinates and radius

    Examples
    --------
    >>> track = track_site_parameters(
    ...     step=30, pmin=950, vmean=10, delta=45, gamma=270,
    ...     x0=-75, y0=25, lon0=-80, lon1=-70, lat0=20, lat1=30,
    ...     R=5, date_ini='2020-09-01 00:00'
    ... )
    >>> print(f"Track points: {len(track)}")
    Track points: 96
    """

    # storm entrance coordinates at the domain boundary
    x1, y1 = entrance_coords(delta, gamma, x0, y0, R, lon0, lon1, lat0, lat1)

    # calculate computational timestep storm coordinates
    # Note: velocity input in shoot function must be km/h
    lon, lat = [x1], [y1]
    i = 1
    glon, glat, baz = shoot(x1, y1, gamma + 180, vmean * 1.852 * i * step / 60)
    if glon < 0:
        glon += 360
    while (glon < lon1) & (glon > lon0) & (glat < lat1) & (glat > lat0):
        lon.append(glon)
        lat.append(glat)
        i += 1
        glon, glat, baz = shoot(x1, y1, gamma + 180, vmean * 1.852 * i * step / 60)
        if glon < 0:
            glon += 360
    frec = len(lon)

    # select Pmin-Vmax polynomial fitting coefficients (IBTrACS center,basin)
    xds_coef = ibtracs_fit_pmin_wmax()
    p1, p2, p3, p4 = xds_coef.sel(center=center, basin=basin).coef.values[:]
    wind_estimate = (
        p1 * np.power(pmin, 3) + p2 * np.power(pmin, 2) + p3 * np.power(pmin, 1) + p4
    )

    # radii of maximum winds is filled with Knaff (2016) estimate
    rmw_estimate = wind2rmw(
        np.full(lat.size, wind_estimate), np.full(lat.size, vmean), lat
    )

    # velocity components
    vfx = vmean * np.sin((gamma + 180) * np.pi / 180)  # [kt]
    vfy = vmean * np.cos((gamma + 180) * np.pi / 180)  # [kt]

    # time array for SWAN input
    time_input = pd.date_range(date_ini, periods=frec, freq="{0}MIN".format(step))

    # storm track (pd.DataFrame)
    st = pd.DataFrame(
        index=time_input,
        columns=["lon", "lat", "move", "vf", "vfx", "vfy", "pn", "p0", "vmax", "rmw"],
    )

    st["lon"] = lon
    st["lat"] = lat
    st["move"] = gamma  # gamma, forward direction
    st["vf"] = vmean  # translation speed [kt]
    st["vfx"] = vfx  # x-component
    st["vfy"] = vfy  # y-component
    st["pn"] = 1013  # average pressure at the surface [mbar]
    st["p0"] = pmin  # minimum central pressure [mbar]
    st["vmax"] = wind_estimate  # maximum winds [kt, 1-min avg]
    st["rmw"] = rmw_estimate  # radii of maximum winds [nmile]

    # add metadata
    st.attrs = {
        "vf": "kt",
        "p0": "mbar",
        "vmax": "kt, 1-min avg",
        "rmw": "nmile",
        "x0": x0,
        "y0": y0,
        "R": 4,
    }

    return st


###############################################################################
# synthetic tracks
###############################################################################


def nakajo_track_preprocessing(xds: xr.Dataset, center: str = "WMO") -> pd.DataFrame:
    """
    Preprocess synthetic storm track data from Nakajo format.

    Parameters
    ----------
    xds : xr.Dataset
        Synthetic storm track dataset with storm dimension in Nakajo format.
        Must contain variables for time, longitude (ylon_TC), latitude (ylat_TC),
        and pressure (yCPRES).
    center : str, optional
        IBTrACS center code. Default is "WMO".

    Returns
    -------
    pd.DataFrame
        Preprocessed storm track data with columns:
        - center : Storm center identifier
        - basin : Storm basin identifier (determined from coordinates)
        - dist2land : Distance to nearest land (fixed at 100 km)
        - longitude : Storm longitude (0-360°)
        - latitude : Storm latitude
        - move : Forward direction (degrees)
        - mean_velocity : Translation speed (kt)
        - pressure : Central pressure (mbar)
        - maxwinds : Maximum winds (kt, not provided)
        - rmw : Radius of maximum winds (not provided)
        - category : Storm category (0-6)
        - timestep : Time step (hours)
        - storm_vmean : Mean translation speed (kt)

    Notes
    -----
    The function:
    1. Removes NaT (Not a Time) values from time coordinate
    2. Removes NaN values from pressure data
    3. Converts longitudes to [0°-360°] convention
    4. Determines basin based on storm coordinates
    5. Rounds dates to nearest hour
    6. Calculates translation speed and direction
    7. Determines storm category based on pressure

    Examples
    --------
    >>> processed = nakajo_track_preprocessing(nakajo_data, center='WMO')
    >>> print(f"Track duration: {len(processed)} time steps")
    Track duration: 48
    >>> print(f"Storm category: {processed['category'].iloc[0]}")
    Storm category: 3
    """

    # get names of variables
    nm_tim = "time"
    nm_lon = "ylon_TC"
    nm_lat = "ylat_TC"
    nm_prs = "yCPRES"

    # get var time
    ytime = xds[nm_tim].values  # dates format: datetime64

    # remove NaTs (time)
    ylat_tc = xds[nm_lat].values[~np.isnat(ytime)]  # latitude
    ylon_tc = xds[nm_lon].values[~np.isnat(ytime)]  # longitude
    ycpres = xds[nm_prs].values[~np.isnat(ytime)]  # pressure [mbar]
    ytime = ytime[~np.isnat(ytime)]

    # remove common NaNs (pressure)
    posnonan_p_w = np.unique(np.argwhere(~np.isnan(ycpres)))
    ytime = ytime[posnonan_p_w]
    st_lat = ylat_tc[posnonan_p_w]
    st_lon = ylon_tc[posnonan_p_w]
    st_pres = ycpres[posnonan_p_w]

    ###########################################################################
    # longitude convention: [0º,360º]
    st_lon[st_lon < 0] = st_lon[st_lon < 0] + 360

    # get basin
    ybasin = np.empty(ytime.size, dtype="S10")
    for i in range(ytime.size):
        ilo, ila = st_lon[i], st_lat[i]
        if (ila < 0) & (ilo < 135) & (ilo > 20):
            ybasin[i] = "SI"
        elif (ila < 0) & (ilo > 135) & (ilo < 290):
            ybasin[i] = "SP"
        elif (ila < 0) & (ilo < 20) & (ilo > 290):
            ybasin[i] = "SA"
        elif (ila > 0) & (ilo > 20) & (ilo < 100):
            ybasin[i] = "NI"
        elif (ila > 0) & (ilo > 100) & (ilo < 180):
            ybasin[i] = "WP"
        elif (ila > 0) & (ilo > 180) & (ilo < 260):
            ybasin[i] = "EP"
        elif (ila > 0) & (ila < 15) & (ilo > 260) & (ilo < 275):
            ybasin[i] = "EP"
        else:
            ybasin[i] = "NA"

    ###########################################################################

    # only when storm data available
    if st_pres.size > 0:
        # round dates to hour
        round_to = 3600
        st_time = []
        for i in range(len(ytime)):
            dt = ytime[i].astype("datetime64[s]").tolist()
            seconds = (dt - dt.min).seconds
            rounding = (seconds + round_to / 2) // round_to * round_to
            out = dt + timedelta(0, rounding - seconds, -dt.microsecond)
            st_time.append(out)
        st_time = np.asarray(st_time)

        # storm coordinates timestep [hours]
        ts = st_time[1:] - st_time[:-1]
        ts = [ts[i].total_seconds() / 3600 for i in range(ts.size)]
        ts.append(np.nan)

        # calculate Vmean
        st_vmean, st_move = [], []
        for i in range(0, len(st_time) - 1):
            # consecutive storm coordinates
            lon1, lon2 = st_lon[i], st_lon[i + 1]
            lat1, lat2 = st_lat[i], st_lat[i + 1]

            # translation speed
            gamma_h, vel_mean, _, _ = get_vmean(lat1, lon1, lat2, lon2, ts[i])
            st_vmean.append(vel_mean / 1.852)  # translation speed [km/h to kt]
            st_move.append(gamma_h)  # forward direction [º]
        st_vmean.append(np.nan)
        st_move.append(np.nan)

        # mean value
        vmean = np.nanmean(st_vmean)  # [kt]

        # storm category
        categ = get_category(st_pres)

        # get basin string
        st_basin = [str(c.decode("UTF-8")) for c in ybasin]

        # store storm variables
        df = pd.DataFrame(
            index=st_time,
            columns=[
                "center",
                "basin",
                "dist2land",
                "longitude",
                "latitude",
                "move",
                "mean_velocity",
                "pressure",
                "maxwinds",
                "rmw",
                "category",
                "timestep",
                "storm_vmean",
            ],
        )

        df["center"] = center  # hypothesis winds 1-min
        df["basin"] = st_basin
        df["dist2land"] = 100
        df["longitude"] = st_lon
        df["latitude"] = st_lat
        df["move"] = st_move
        df["mean_velocity"] = st_vmean
        df["pressure"] = st_pres
        df["maxwinds"] = np.nan
        df["rmw"] = np.nan
        df["category"] = categ
        df["timestep"] = ts
        df["storm_vmean"] = vmean

        df.attrs = {
            "dist2land": "km",
            "velocity": "kt",
            "pressure": "mbar",
            "maxwinds": "kt, 1-min average sustained winds (converted from X-min)",
            "rmw": "nmile",
            "timestep": "hour",
        }

    else:
        df = pd.DataFrame(
            index=[],
            columns=[
                "center",
                "basin",
                "dist2land",
                "longitude",
                "latitude",
                "move",
                "mean_velocity",
                "pressure",
                "maxwinds",
                "rmw",
                "category",
                "timestep",
                "storm_vmean",
            ],
        )

    return df



import unittest

import numpy as np

from bluemath_tk.topo_bathy.mesh_utils import (
    calculate_edges,
    detect_circumcenter_too_close,
    get_raster_resolution_meters,
)


class TestDetectCircumcenterTooClose(unittest.TestCase):
    """Test the detect_circumcenter_too_close function."""

    def test_function(self):
        """Test the function with a simple case."""
        # Define the input arrays
        X = np.array([0, 1, 0, 1, 2, 3, 3])
        Y = np.array([0, 0, 1, 1, 1, 0, 1])
        elements = np.array([[0, 1, 2], [1, 3, 2], [3, 4, 5], [4, 5, 6]])
        aj_threshold = 0.1

        # Call the function
        bad_elements_mask = detect_circumcenter_too_close(
            X=X, Y=Y, elements=elements, aj_threshold=aj_threshold
        )

        # Check the result
        expected_mask = np.array([True, True, False, False])
        np.testing.assert_array_equal(bad_elements_mask, expected_mask)


class TestCalculateEdges(unittest.TestCase):
    """Test the calculate_edges function."""

    def test_function(self):
        """Test the function with a simple case."""
        # Define the input arrays
        elements = np.array([[0, 1, 2], [1, 2, 3], [2, 3, 0]])

        # Call the function
        edges = calculate_edges(elements)

        # Check the result
        expected_edges = np.array([[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]])
        np.testing.assert_array_equal(edges, expected_edges)


class TestGetRasterResolutionMeters(unittest.TestCase):
    """Test the get_raster_resolution_meters function."""

    def test_function(self):
        """Test the function with a simple case."""

        lon_center = 0
        lat_center = 0
        raster_resolution = 0.1
        project = lambda x, y: (x * 100000, y * 100000)

        resolution = get_raster_resolution_meters(
            lon_center=lon_center,
            lat_center=lat_center,
            raster_resolution=raster_resolution,
            project=project,
        )

        # Check the result
        expected_resolution = np.float64(7071.067811865475)
        np.testing.assert_almost_equal(resolution, expected_resolution, decimal=5)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pip
import numpy as np
import matplotlib.colors as mcolors


class MidpointNormalize(mcolors.Normalize):
	'''
	Normalise the colorbar so that diverging bars work there way either side from a prescribed midpoint value)
	e.g. im=ax1.imshow(array, norm=MidpointNormalize(midpoint=0.,vmin=-100, vmax=100))
	'''

	def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
		self.midpoint = midpoint
		mcolors.Normalize.__init__(self, vmin, vmax, clip)

	def __call__(self, value, clip=None):
		x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
		return np.ma.masked_array(np.interp(value, x, y), np.isnan(value))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pip
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.colors as mcolors

# teslakit
from .custom_colors import colors_mjo

# import constants
from .config import _faspect, _fsize, _fdpi


def Plot_MJO_phases(rmm1, rmm2, phase, show=True):
    'Plot MJO data separated by phase'

    # parameters for custom plot
    size_points = 0.2
    size_lines = 0.8
    l_colors_phase = np.array(
        [
            [1, 0, 0],
            [0.6602, 0.6602, 0.6602],
            [1.0, 0.4961, 0.3125],
            [0, 1, 0],
            [0.2539, 0.4102, 0.8789],
            [0, 1, 1],
            [1, 0.8398, 0],
            [0.2930, 0, 0.5078]
        ]
    )

    color_lines_1 = (0.4102, 0.4102, 0.4102)


    # plot figure
    fig, ax = plt.subplots(1,1, figsize=(_fsize, _fsize))
    ax.scatter(rmm1, rmm2, c='b', s=size_points)

    # plot data by phases
    for i in range(1,9):
        ax.scatter(
            rmm1.where(phase==i),
            rmm2.where(phase==i),
            c=np.array([l_colors_phase[i-1]]),
            s=size_points)

    # plot sectors
    ax.plot([-4,4],[-4,4], color='k', linewidth=size_lines)
    ax.plot([-4,4],[4,-4], color='k', linewidth=size_lines)
    ax.plot([-4,4],[0,0],  color='k', linewidth=size_lines)
    ax.plot([0,0], [-4,4], color='k', linewidth=size_lines)

    # axis
    plt.xlim(-4, 4)
    plt.ylim(-4, 4)
    plt.xlabel('RMM1')
    plt.ylabel('RMM2')
    ax.set_aspect('equal')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_MJO_Categories(rmm1, rmm2, categ, show=True):
    'Plot MJO data separated by 25 categories'

    # parameters for custom plot
    size_lines = 0.8
    color_lines_1 = (0.4102, 0.4102, 0.4102)

    # custom colors for mjo 25 categories
    np_colors_rgb_categ = colors_mjo()

    # plot figure
    fig, ax = plt.subplots(1,1, figsize=(_fsize,_fsize))

    # plot sectors
    ax.plot([-4,4],[-4,4], color='k', linewidth=size_lines, zorder=9)
    ax.plot([-4,4],[4,-4], color='k', linewidth=size_lines, zorder=9)
    ax.plot([-4,4],[0,0],  color='k', linewidth=size_lines, zorder=9)
    ax.plot([0,0], [-4,4], color='k', linewidth=size_lines, zorder=9)

    # plot circles
    R = [1, 1.5, 2.5]

    for rr in R:
        ax.add_patch(
            patches.Circle(
                (0,0),
                rr,
                color='k',
                linewidth=size_lines,
                fill=False,
                zorder=9)
        )
    ax.add_patch(
        patches.Circle((0,0),R[0],fc='w',fill=True, zorder=10))

    # plot data by categories
    for i in range(1,25):
        if i>8:
            size_points = 0.2
        else:
            size_points = 1.7

        ax.scatter(
            rmm1.where(categ==i),
            rmm2.where(categ==i),
            c=[np_colors_rgb_categ[i-1]],
            s=size_points
        )

    # last category on top (zorder)
    ax.scatter(
        rmm1.where(categ==25),
        rmm2.where(categ==25),
        c=[np_colors_rgb_categ[-1]],
        s=0.2,
        zorder=11
    )

    # TODO: category number
    rr = 0.3
    ru = 0.2
    l_pn = [
        (-3, -1.5, '1'),
        (-1.5, -3, '2'),
        (1.5-rr, -3, '3'),
        (3-rr, -1.5, '4'),
        (3-rr, 1.5-ru, '5'),
        (1.5-rr, 3-ru, '6'),
        (-1.5, 3-ru, '7'),
        (-3, 1.5-ru, '8'),
        (-2, -1, '9'),
        (-1, -2, '10'),
        (1-rr, -2, '11'),
        (2-rr, -1, '12'),
        (2-rr, 1-ru, '13'),
        (1-rr, 2-ru, '14'),
        (-1, 2-ru, '15'),
        (-2, 1-ru, '16'),
        (-1.3, -0.6, '17'),
        (-0.6, -1.3, '18'),
        (0.6-rr, -1.3, '19'),
        (1.3-rr, -0.6, '20'),
        (1.3-rr, 0.6-ru, '21'),
        (0.6-rr, 1.3-ru, '22'),
        (-0.6, 1.3-ru, '23'),
        (-1.3, 0.6-ru, '24'),
        (0-rr/2, 0-ru/2, '25'),
    ]
    for xt, yt, tt in l_pn:
        ax.text(xt, yt, tt, fontsize=15, fontweight='bold', zorder=11)

    # axis
    plt.xlim(-4, 4)
    plt.ylim(-4, 4)
    plt.xlabel('RMM1')
    plt.ylabel('RMM2')
    ax.set_aspect('equal')

    # show and return figure
    if show: plt.show()
    return fig




from typing import Dict, List, Tuple, Union

import numpy as np


def energy_spectrum(hs: float, tp: float, gamma: float, duration: int) -> List[float]:
    """
    Calculate JONSWAP energy spectrum for given wave parameters.

    This function computes the energy spectrum S(f) using the JONSWAP spectral model,
    which is commonly used for wind-generated waves in deep water.

    Parameters
    ----------
    hs : float
        Significant wave height in meters.
    tp : float
        Peak period in seconds.
    gamma : float
        JONSWAP peak enhancement factor (typically 1-7, default 3.3).
    duration : int
        Number of frequency points for the spectrum.

    Returns
    -------
    List[float]
        Energy spectrum values S(f) for each frequency.

    Notes
    -----
    The JONSWAP spectrum combines the Pierson-Moskowitz spectrum with a peak
    enhancement factor. The spectrum is defined as:

    S(f) = Beta * (Hs^2) * (Tp^-4) * (f^-5) * exp(-1.25 * (Tp*f)^-4) * gamma^exp(-((Tp*f-1)^2)/(2*sigma^2))

    where sigma varies based on frequency relative to peak frequency.
    """

    # Defining frequency series - tend length
    freqs = np.linspace(0.02, 1, duration)

    S = []
    fp = 1 / tp

    for f in freqs:
        if f <= fp:
            sigma = 0.07
        if f > fp:
            sigma = 0.09

        Beta = (0.06238 / (0.23 + 0.0336 * gamma - 0.185 * (1.9 + gamma) ** -1)) * (
            1.094 - 0.01915 * np.log(gamma)
        )
        Sf = (
            Beta
            * (hs**2)
            * (tp**-4)
            * (f**-5)
            * np.exp(-1.25 * (tp * f) ** -4)
            * gamma ** (np.exp((-((tp * f - 1) ** 2)) / (2 * sigma**2)))
        )
        S.append(Sf)

    return S


def series_Jonswap(waves: Dict[str, Union[float, int]]) -> np.ndarray:
    """
    Generate surface elevation time series from JONSWAP spectrum.

    This function generates a time series of surface elevation using the JONSWAP
    spectral model. The elevation is computed by summing harmonic components
    with random phases.

    Parameters
    ----------
    waves : Dict[str, Union[float, int]]
        Dictionary containing wave parameters:
        - 'H': Significant wave height (m)
        - 'T': Peak period (s)
        - 'gamma': JONSWAP peak enhancement factor
        - 'warmup': Spin-up time (s)
        - 'deltat': Time step (s)
        - 'tendc': Simulation period (s)

    Returns
    -------
    np.ndarray
        2D array with shape (n_time_steps, 2) containing:
        - Column 0: Time values (s)
        - Column 1: Surface elevation values (m)

    Notes
    -----
    The surface elevation is generated by:
    1. Computing the JONSWAP energy spectrum
    2. Converting spectrum to amplitude components
    3. Summing harmonic components with random phases
    """

    # waves properties
    hs = waves["H"]
    tp = waves["T"]
    gamma = waves["gamma"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    duration = int(tendc + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    # why? wee user manual
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate energy spectrum
    S = energy_spectrum(hs, tp, gamma, duration)

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_TMA(waves: Dict[str, Union[float, int]], depth: float) -> np.ndarray:
    """
    Generate surface elevation time series from TMA spectrum.

    This function generates a time series of surface elevation using the TMA
    spectral model, which extends the JONSWAP spectrum to include depth effects.

    Parameters
    ----------
    waves : Dict[str, Union[float, int]]
        Dictionary containing wave parameters:
        - 'H': Significant wave height (m)
        - 'T': Peak period (s)
        - 'gamma': JONSWAP peak enhancement factor
        - 'warmup': Spin-up time (s)
        - 'deltat': Time step (s)
        - 'comptime': Simulation period (s)
    depth : float
        Water depth in meters.

    Returns
    -------
    np.ndarray
        2D array with shape (n_time_steps, 2) containing:
        - Column 0: Time values (s)
        - Column 1: Surface elevation values (m)

    Notes
    -----
    The TMA spectrum modifies the JONSWAP spectrum to account for depth effects
    using a depth function fi that depends on wave number and depth.
    """

    # waves properties
    hs = waves["H"]
    tp = waves["T"]
    gamma = waves["gamma"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    comptime = waves["comptime"]
    _g = 9.801

    # series duration
    duration = int(comptime + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    # why? wee user manual
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate JONSWAP energy spectrum
    S = energy_spectrum(hs, tp, gamma, duration)
    m0_Jonswap = np.trapz(S, x=freqs)

    # Transform JONSWAP into TMA by (omega, h)
    S_TMA = []
    for pf, f in enumerate(freqs):
        """
        omega = 2 * np.pi * f
        
        thv = omega * np.sqrt(depth / g)
        if thv < 1: 
            fi = 0.5 * (thv)**2
        elif (thv >= 1) and (thv < 2):
            fi = 1 - 0.5 * (2 - thv)**2
        else:
            fi = 1
        """
        # S_TMA.append(S[pf] * fi)

        # As in Holthuijsen
        L = waves_dispersion(1 / f, depth)[0]
        k = 2 * np.pi / L
        n = 0.5 * (1 + ((2 * k * depth) / (np.sinh(2 * k * depth))))
        fi = (1 / (2 * n)) * np.tanh(k * depth)

        S_TMA.append(S[pf] * fi)

    m0_TMA = np.trapz(S_TMA, x=freqs)

    S_TMA_mod = [i * (m0_Jonswap / m0_TMA) for i in S_TMA]

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S_TMA_mod[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_TMA_bimodal(waves: Dict[str, Union[float, int]], depth: float) -> np.ndarray:
    """
    Generate surface elevation time series from bimodal TMA spectrum.

    This function generates a time series of surface elevation using a bimodal
    TMA spectral model, combining two wave systems with different characteristics.

    Parameters
    ----------
    waves : Dict[str, Union[float, int]]
        Dictionary containing wave parameters for two wave systems:
        - 'Hs1': Significant wave height of system 1 (m)
        - 'Tp1': Peak period of system 1 (s)
        - 'gamma1': JONSWAP peak enhancement factor for system 1
        - 'Hs2': Significant wave height of system 2 (m)
        - 'Tp2': Peak period of system 2 (s)
        - 'gamma2': JONSWAP peak enhancement factor for system 2
        - 'warmup': Spin-up time (s)
        - 'deltat': Time step (s)
        - 'tendc': Simulation period (s)
    depth : float
        Water depth in meters.

    Returns
    -------
    np.ndarray
        2D array with shape (n_time_steps, 2) containing:
        - Column 0: Time values (s)
        - Column 1: Surface elevation values (m)

    Notes
    -----
    The bimodal spectrum is created by summing two TMA spectra, each representing
    a different wave system (e.g., wind sea and swell).
    """

    # waves properties
    hs1 = waves["Hs1"]
    tp1 = waves["Tp1"]
    gamma1 = waves["gamma1"]
    hs2 = waves["Hs2"]
    tp2 = waves["Tp2"]
    gamma2 = waves["gamma2"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    # TODO: puede que haya algun problema en esta suma
    duration = int(tendc + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate energy spectrum for each wave system
    S1 = energy_spectrum(hs1, tp1, gamma1, duration)
    S2 = energy_spectrum(hs2, tp2, gamma2, duration)

    m01_Jonswap = np.trapz(S1, x=freqs)
    m02_Jonswap = np.trapz(S2, x=freqs)

    # Transform JONSWAP into TMA by (omega, h)
    S_TMA_1, S_TMA_2 = [], []
    for pf, f in enumerate(freqs):
        # As in Holthuijsen
        L = waves_dispersion(1 / f, depth)[0]
        k = 2 * np.pi / L
        n = 0.5 * (1 + ((2 * k * depth) / (np.sinh(2 * k * depth))))
        fi = (1 / (2 * n)) * np.tanh(k * depth)

        S_TMA_1.append(S1[pf] * fi)
        S_TMA_2.append(S2[pf] * fi)

    m0_TMA_1 = np.trapz(S_TMA_1, x=freqs)
    m0_TMA_2 = np.trapz(S_TMA_2, x=freqs)

    S_TMA_mod_1 = [i * (m01_Jonswap / m0_TMA_1) for i in S_TMA_1]
    S_TMA_mod_2 = [i * (m02_Jonswap / m0_TMA_2) for i in S_TMA_2]

    # bimodal wave spectra
    S = np.sum([S_TMA_mod_1, S_TMA_mod_2], axis=0)

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_Jonswap_bimodal(waves: Dict[str, Union[float, int]]) -> np.ndarray:
    """
    Generate surface elevation time series from bimodal JONSWAP spectrum.

    This function generates a time series of surface elevation using a bimodal
    JONSWAP spectral model, combining two wave systems with different characteristics.

    Parameters
    ----------
    waves : Dict[str, Union[float, int]]
        Dictionary containing wave parameters for two wave systems:
        - 'Hs1': Significant wave height of system 1 (m)
        - 'Tp1': Peak period of system 1 (s)
        - 'gamma1': JONSWAP peak enhancement factor for system 1
        - 'Hs2': Significant wave height of system 2 (m)
        - 'Tp2': Peak period of system 2 (s)
        - 'gamma2': JONSWAP peak enhancement factor for system 2
        - 'warmup': Spin-up time (s)
        - 'deltat': Time step (s)
        - 'tendc': Simulation period (s)

    Returns
    -------
    np.ndarray
        2D array with shape (n_time_steps, 2) containing:
        - Column 0: Time values (s)
        - Column 1: Surface elevation values (m)

    Notes
    -----
    The bimodal spectrum is created by summing two JONSWAP spectra, each representing
    a different wave system (e.g., wind sea and swell).
    """

    # waves properties
    hs1 = waves["Hs1"]
    tp1 = waves["Tp1"]
    gamma1 = waves["gamma1"]
    hs2 = waves["Hs2"]
    tp2 = waves["Tp2"]
    gamma2 = waves["gamma2"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    # TODO: puede que haya algun problema en esta suma
    duration = int(tendc + warmup)
    time = np.arange(0, duration, deltat)

    # series frequency
    freqs = np.linspace(0.02, 1, duration)
    delta_f = freqs[1] - freqs[0]

    # calculate energy spectrum for each wave system
    S1 = energy_spectrum(hs1, tp1, gamma1, duration)
    S2 = energy_spectrum(hs2, tp2, gamma2, duration)
    # bimodal wave spectra
    S = np.sum([S1, S2], axis=0)

    # series elevation
    teta = np.zeros((len(time)))

    # calculate aij
    for f in range(len(freqs)):
        ai = np.sqrt(S[f] * 2 * delta_f)
        eps = np.random.rand() * (2 * np.pi)

        # calculate elevation
        teta = teta + ai * np.cos(2 * np.pi * freqs[f] * time + eps)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_regular_monochromatic(waves: Dict[str, Union[float, int]]) -> np.ndarray:
    """
    Generate monochromatic regular wave time series.

    This function generates a time series of surface elevation for monochromatic
    regular waves with constant amplitude and frequency.

    Parameters
    ----------
    waves : Dict[str, Union[float, int]]
        Dictionary containing wave parameters:
        - 'T': Wave period (s)
        - 'H': Wave height (m)
        - 'warmup': Spin-up time (s)
        - 'deltat': Time step (s)
        - 'comptime': Simulation period (s)

    Returns
    -------
    np.ndarray
        2D array with shape (n_time_steps, 2) containing:
        - Column 0: Time values (s)
        - Column 1: Surface elevation values (m)

    Notes
    -----
    The surface elevation follows a simple cosine function:
    η(t) = (H/2) * cos(2πt/T)
    """

    # waves properties
    T = waves["T"]
    H = waves["H"]
    # WL = waves["WL"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["comptime"]

    # series duration
    duration = tendc + int(warmup)
    time = np.arange(0, duration, deltat)

    # series elevation
    teta = (H / 2) * np.cos((2 * np.pi / T) * time)

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def series_regular_bichromatic(waves: Dict[str, Union[float, int]]) -> np.ndarray:
    """
    Generate bichromatic regular wave time series.

    This function generates a time series of surface elevation for bichromatic
    regular waves, combining two monochromatic components.

    Parameters
    ----------
    waves : Dict[str, Union[float, int]]
        Dictionary containing wave parameters:
        - 'T1': Period of component 1 (s)
        - 'T2': Period of component 2 (s)
        - 'H': Wave height (m)
        - 'WL': Water level (m)
        - 'warmup': Spin-up time (s)
        - 'deltat': Time step (s)
        - 'tendc': Simulation period (s)

    Returns
    -------
    np.ndarray
        2D array with shape (n_time_steps, 2) containing:
        - Column 0: Time values (s)
        - Column 1: Surface elevation values (m)

    Notes
    -----
    The surface elevation is the sum of two cosine components:
    η(t) = (H/2) * cos(2πt/T1) + (H/2) * cos(2πt/T2)
    """

    # waves properties
    T1 = waves["T1"]
    T2 = waves["T2"]
    H = waves["H"]
    _WL = waves["WL"]
    warmup = waves["warmup"]
    deltat = waves["deltat"]
    tendc = waves["tendc"]

    # series duration
    duration = tendc + int(warmup)
    time = np.arange(0, duration, deltat)

    # series elevation
    teta = (H / 2) * np.cos((2 * np.pi / T1) * time) + (H / 2) * np.cos(
        (2 * np.pi / T2) * time
    )

    # generate series dataframe
    series = np.zeros((len(time), 2))
    series[:, 0] = time
    series[:, 1] = teta

    return series


def waves_dispersion(T: float, h: float) -> Tuple[float, float, float]:
    """
    Solve the wave dispersion relation to calculate wavelength, wave number, and phase speed.

    This function solves the linear wave dispersion relation to determine wave
    properties in water of finite depth.

    Parameters
    ----------
    T : float
        Wave period in seconds.
    h : float
        Water depth in meters.

    Returns
    -------
    Tuple[float, float, float]
        A tuple containing:
        - L: Wavelength in meters
        - k: Wave number in radians per meter
        - c: Phase speed in meters per second

    Notes
    -----
    The dispersion relation is solved iteratively using:
    L = (g*T²)/(2π) * tanh(2πh/L)

    where g is gravitational acceleration (9.81 m/s²).

    The wave number k = 2π/L and phase speed c = √(g*tanh(kh)/k).
    """

    L1 = 1
    L2 = ((9.81 * T**2) / 2 * np.pi) * np.tanh(h * 2 * np.pi / L1)
    umbral = 1

    while umbral > 0.1:
        L2 = ((9.81 * T**2) / (2 * np.pi)) * np.tanh((h * 2 * np.pi) / L1)
        umbral = np.abs(L2 - L1)
        L1 = L2

    L = L2
    k = (2 * np.pi) / L
    c = np.sqrt(9.8 * np.tanh(k * h) / k)

    return (L, k, c)



import math
from io import BytesIO
from typing import Tuple

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.pyplot as plt
import requests
from PIL import Image

from ..constants import EARTH_RADIUS

EARTH_RADIUS_M = EARTH_RADIUS * 1000  # Convert km to m


def deg2num(lat: float, lon: float, zoom: int) -> Tuple[int, int]:
    """
    Converts geographic coordinates to tile numbers for a given zoom level.

    Parameters
    ----------
    lat : float
        Latitude in degrees.
    lon : float
        Longitude in degrees.
    zoom : int
        Zoom level.

    Returns
    -------
    xtile : int
        Tile number in x-direction.
    ytile : int
        Tile number in y-direction.
    """

    lat_rad = math.radians(lat)
    n = 2.0**zoom
    xtile = int((lon + 180.0) / 360.0 * n)
    ytile = int(
        (1.0 - math.log(math.tan(lat_rad) + 1 / math.cos(lat_rad)) / math.pi) / 2.0 * n
    )

    return xtile, ytile


def num2deg(xtile: int, ytile: int, zoom: int) -> Tuple[float, float]:
    """
    Converts tile numbers back to geographic coordinates.

    Parameters
    ----------
    xtile : int
        Tile number in x-direction.
    ytile : int
        Tile number in y-direction.
    zoom : int
        Zoom level.

    Returns
    -------
    lat : float
        Latitude in degrees.
    lon : float
        Longitude in degrees.
    """

    n = 2.0**zoom
    lon = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat = math.degrees(lat_rad)

    return lat, lon


def lonlat_to_webmercator(lon: float, lat: float) -> Tuple[float, float]:
    """
    Converts lon/lat to Web Mercator projection coordinates in meters.

    Parameters
    ----------
    lon : float
        Longitude in degrees.
    lat : float
        Latitude in degrees.

    Returns
    -------
    x : float
        X coordinate in meters.
    y : float
        Y coordinate in meters.
    """

    x = EARTH_RADIUS_M * math.radians(lon)
    y = EARTH_RADIUS_M * math.log(math.tan((math.pi / 4) + math.radians(lat) / 2))

    return x, y


def tile_bounds_meters(
    x_start: int, y_start: int, x_end: int, y_end: int, zoom: int
) -> Tuple[float, float, float, float]:
    """
    Computes the bounding box of the tile region in Web Mercator meters.

    Returns
    -------
    xmin, ymin, xmax, ymax : float
        Bounding box in meters (Web Mercator projection).
    """

    lat1, lon1 = num2deg(x_start, y_start, zoom)
    lat2, lon2 = num2deg(x_end + 1, y_end + 1, zoom)
    x1, y1 = lonlat_to_webmercator(lon1, lat2)
    x2, y2 = lonlat_to_webmercator(lon2, lat1)

    return x1, y1, x2, y2


def calculate_zoom(
    lon_min: float, lon_max: float, display_width_px: int = 1024, tile_size: int = 256
) -> int:
    """
    Automatically estimates an appropriate zoom level for the bounding box.

    Returns
    -------
    zoom : int
        Estimated zoom level.
    """

    WORLD_MAP_WIDTH = 2 * math.pi * EARTH_RADIUS_M
    x1, _ = lonlat_to_webmercator(lon_min, 0)
    x2, _ = lonlat_to_webmercator(lon_max, 0)
    region_width_m = abs(x2 - x1)
    meters_per_pixel_desired = region_width_m / display_width_px
    zoom = math.log2(WORLD_MAP_WIDTH / (tile_size * meters_per_pixel_desired))

    return int(round(zoom))


def get_cartopy_scale(zoom: int) -> str:
    """
    Select appropriate cartopy feature scale based on zoom level.

    Parameters
    ----------
    zoom : int
        Web Mercator zoom level.

    Returns
    -------
    scale : str
        One of '110m', '50m', or '10m'.
    """

    if zoom >= 9:
        return "10m"
    elif zoom >= 6:
        return "50m"
    else:
        return "110m"


def plot_usgs_raster_map(
    lat_min: float,
    lat_max: float,
    lon_min: float,
    lon_max: float,
    zoom: int = None,
    source: str = "arcgis",
    verbose: bool = True,
    mask_ocean: bool = False,
    add_features: bool = True,
    display_width_px: int = 1024,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Downloads and displays a USGS raster map for the given bounding box.

    Parameters
    ----------
    lat_min : float
        Minimum latitude of the bounding box.
    lat_max : float
        Maximum latitude of the bounding box.
    lon_min : float
        Minimum longitude of the bounding box.
    lon_max : float
        Maximum longitude of the bounding box.
    zoom : int, optional
        Zoom level (default is None, which auto-calculates based on bounding box).
    source : str, optional
        Source of the map tiles (default is "arcgis"). Options include:
        - "arcgis" (Esri World Imagery)
        - "google" (Google Maps Satellite)
        - "eox" (EOX Sentinel-2 Cloudless)
        - "osm" (OpenStreetMap Standard Tiles)
        - "amazon" (Amazon Terrarium Elevation Tiles)
        - "esri_world" (Esri World Imagery)
        - "geoportail_fr" (Geoportail France Orthophotos)
        - "ign_spain_pnoa" (IGN Spain PNOA Orthophotos)
    verbose : bool, optional
        If True, prints additional information about the source and zoom level.
    mask_ocean : bool, optional
        If True, masks ocean areas in the map.
    add_features : bool, optional
        If True, adds borders, coastlines, and states to the map.
    display_width_px : int, optional
        Width of the display in pixels (default is 1024).

    Returns
    -------
    fig : plt.Figure
        The matplotlib figure containing the map.
    ax : plt.Axes
        The matplotlib axes with the map projection.
    """

    tile_size = 256
    if zoom is None:
        zoom = calculate_zoom(lon_min, lon_max, display_width_px, tile_size)
    if verbose:
        print(f"Auto-selected zoom level: {zoom}")

    x_start, y_start = deg2num(lat_max, lon_min, zoom)
    x_end, y_end = deg2num(lat_min, lon_max, zoom)
    width = x_end - x_start + 1
    height = y_end - y_start + 1

    map_img = Image.new("RGB", (width * tile_size, height * tile_size))
    if source == "arcgis":
        tile_url = "https://services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}"
        z_max = 19
        if verbose:
            print(
                "Using Esri World Imagery (ArcGIS):\n"
                "- Free for public and non-commercial use\n"
                "- Commercial use allowed under Esri terms of service\n"
                "- Attribution required: 'Tiles © Esri — Sources: Esri, Earthstar Geographics, CNES/Airbus DS, "
                "USDA, USGS, AeroGRID, IGN, and the GIS User Community'\n"
                "- Max zoom ~19"
            )

    elif source == "google":
        tile_url = "https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}"
        z_max = 21
        if verbose:
            print(
                "Using Google Maps Satellite:\n"
                "- NOT license-free\n"
                "- Usage outside official Google Maps SDKs/APIs is prohibited\n"
                "- Commercial use requires a paid license from Google\n"
                "- May be blocked without an API key\n"
                "- Max zoom ~21"
            )

    elif source == "eox":
        tile_url = "https://tiles.maps.eox.at/wmts/1.0.0/s2cloudless-2024_3857/default/g/{z}/{y}/{x}.jpg"
        z_max = 16
        if verbose:
            print(
                "Using EOX Sentinel-2 Cloudless:\n"
                "- Based on Copernicus Sentinel-2 data processed by EOX\n"
                "- Licensed under Creative Commons BY-NC-SA 4.0 (non-commercial use only)\n"
                "- Attribution required: 'Sentinel-2 cloudless – © EOX, based on modified Copernicus Sentinel data 2016–2024'\n"
                "- Versions from 2016–2017 are under CC BY 4.0 (commercial use allowed)\n"
                "- Max zoom ~16"
            )

    elif source == "osm":
        tile_url = "https://tile.openstreetmap.org/{z}/{x}/{y}.png"
        z_max = 19
        if verbose:
            print(
                "Using OpenStreetMap Standard Tiles:\n"
                "- Free and open-source (ODbL license)\n"
                "- Commercial use allowed with attribution\n"
                "- Attribution required: '© OpenStreetMap contributors'\n"
                "- Max zoom ~19"
            )

    elif source == "amazon":
        tile_url = (
            "https://s3.amazonaws.com/elevation-tiles-prod/terrarium/{z}/{x}/{y}.png"
        )
        z_max = 15
        if verbose:
            print(
                "Using Amazon Terrarium Elevation Tiles:\n"
                "- Free for public use\n"
                "- Attribution recommended: 'Amazon Terrarium'\n"
                "- Max zoom ~15"
            )

    elif source == "esri_world":
        tile_url = "https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}"
        z_max = 19
        if verbose:
            print(
                "Using Esri World Imagery:\n"
                "- High-resolution global imagery\n"
                "- Free with attribution under Esri terms\n"
                "- Max zoom ~19"
            )

    elif source == "geoportail_fr":
        tile_url = (
            "https://data.geopf.fr/wmts?"
            "REQUEST=GetTile&SERVICE=WMTS&VERSION=1.0.0"
            "&STYLE=normal&TILEMATRIXSET=PM&FORMAT=image/jpeg"
            "&LAYER=ORTHOIMAGERY.ORTHOPHOTOS&TILEMATRIX={z}"
            "&TILEROW={y}&TILECOL={x}"
        )
        z_max = 19
        if verbose:
            print(
                "Using Geoportail France (Orthophotos):\n"
                "- Aerial orthophotos from the French National Institute (IGN)\n"
                "- Free for public use under Etalab license\n"
                "- Attribution: Geoportail France / IGN\n"
                "- Max zoom ~19"
            )
    elif source == "ign_spain_pnoa":
        tile_url = "https://tms-pnoa-ma.idee.es/1.0.0/pnoa-ma/{z}/{x}/{y_inv}.jpeg"
        z_max = 19
        if verbose:
            print(
                "Using IGN Spain PNOA Orthophotos:\n"
                "- High-resolution aerial imagery (PNOA program)\n"
                "- Provided by IGN/CNIG (Government of Spain)\n"
                "- Free to use under Creative Commons BY 4.0 license\n"
                "- Attribution required: 'Ortofotografía PNOA – © IGN / CNIG (Gobierno de España) – CC BY 4.0'\n"
                "- Max zoom ~19"
            )

    if zoom > z_max:
        zoom = z_max

    for x in range(x_start, x_end + 1):
        for y in range(y_start, y_end + 1):
            if source == "ign_spain_pnoa":
                y_inv = (2**zoom - 1) - y
                url = tile_url.format(z=zoom, x=x, y_inv=y_inv)
            else:
                url = tile_url.format(z=zoom, x=x, y=y)
            try:
                response = requests.get(url, timeout=10)
                tile = Image.open(BytesIO(response.content))
                map_img.paste(
                    tile, ((x - x_start) * tile_size, (y - y_start) * tile_size)
                )
            except Exception as e:
                print(f"Error fetching tile {x},{y}: {e}")

    xmin, ymin, xmax, ymax = tile_bounds_meters(x_start, y_start, x_end, y_end, zoom)

    crs_tiles = ccrs.PlateCarree()
    fig = plt.figure(figsize=(12, 8))
    ax = plt.axes(projection=crs_tiles)
    ax.set_extent([lon_min, lon_max, lat_min, lat_max], crs=crs_tiles)
    ax.imshow(
        map_img,
        origin="upper",
        extent=[xmin, xmax, ymin, ymax],
        transform=ccrs.Mercator.GOOGLE,
    )
    scale = get_cartopy_scale(zoom)
    if verbose:
        print(f"Using Cartopy scale: {scale}")

    if add_features:
        ax.add_feature(cfeature.BORDERS.with_scale(scale), linewidth=0.8)
        ax.add_feature(cfeature.COASTLINE.with_scale(scale))
        ax.add_feature(cfeature.STATES.with_scale(scale), linewidth=0.5)

    if mask_ocean:
        ax.add_feature(cfeature.OCEAN.with_scale(scale), facecolor="w", zorder=1)

    return fig, ax


if __name__ == "__main__":
    lat_min, lat_max = 35.406, 54.372
    lon_min, lon_max = -13.148, 12.325
    fig, ax = plot_usgs_raster_map(
        lat_min,
        lat_max,
        lon_min,
        lon_max,
        zoom=None,
        verbose=True,
        add_features=False,
        mask_ocean=True,
    )
    fig.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.colors as mcolors
from matplotlib import cm

def colors_awt():

    # 6 AWT colors
    l_colors_dwt = [
        (155/255.0, 0, 0),
        (1, 0, 0),
        (255/255.0, 216/255.0, 181/255.0),
        (164/255.0, 226/255.0, 231/255.0),
        (0/255.0, 190/255.0, 255/255.0),
        (51/255.0, 0/255.0, 207/255.0),
    ]

    return np.array(l_colors_dwt)

def colors_mjo():
    'custom colors for MJO 25 categories'

    l_named_colors = [
        'lightskyblue', 'deepskyblue', 'royalblue', 'mediumblue',
        'darkblue', 'darkblue', 'darkturquoise', 'turquoise',
        'maroon', 'saddlebrown', 'chocolate', 'gold', 'orange',
        'orangered', 'red', 'firebrick', 'Purple', 'darkorchid',
        'mediumorchid', 'magenta', 'mediumslateblue', 'blueviolet',
        'darkslateblue', 'indigo', 'darkgray',
    ]

    # get rgb colors as numpy array
    np_colors_rgb = np.array(
        [mcolors.to_rgb(c) for c in l_named_colors]
    )

    return np_colors_rgb

def colors_dwt(num_clusters):

    # 42 DWT colors
    l_colors_dwt = [
        (1.0000, 0.1344, 0.0021),
        (1.0000, 0.2669, 0.0022),
        (1.0000, 0.5317, 0.0024),
        (1.0000, 0.6641, 0.0025),
        (1.0000, 0.9287, 0.0028),
        (0.9430, 1.0000, 0.0029),
        (0.6785, 1.0000, 0.0031),
        (0.5463, 1.0000, 0.0032),
        (0.2821, 1.0000, 0.0035),
        (0.1500, 1.0000, 0.0036),
        (0.0038, 1.0000, 0.1217),
        (0.0039, 1.0000, 0.2539),
        (0.0039, 1.0000, 0.4901),
        (0.0039, 1.0000, 0.6082),
        (0.0039, 1.0000, 0.8444),
        (0.0039, 1.0000, 0.9625),
        (0.0039, 0.8052, 1.0000),
        (0.0039, 0.6872, 1.0000),
        (0.0040, 0.4510, 1.0000),
        (0.0040, 0.3329, 1.0000),
        (0.0040, 0.0967, 1.0000),
        (0.1474, 0.0040, 1.0000),
        (0.2655, 0.0040, 1.0000),
        (0.5017, 0.0040, 1.0000),
        (0.6198, 0.0040, 1.0000),
        (0.7965, 0.0040, 1.0000),
        (0.8848, 0.0040, 1.0000),
        (1.0000, 0.0040, 0.9424),
        (1.0000, 0.0040, 0.8541),
        (1.0000, 0.0040, 0.6774),
        (1.0000, 0.0040, 0.5890),
        (1.0000, 0.0040, 0.4124),
        (1.0000, 0.0040, 0.3240),
        (1.0000, 0.0040, 0.1473),
        (0.9190, 0.1564, 0.2476),
        (0.7529, 0.3782, 0.4051),
        (0.6699, 0.4477, 0.4584),
        (0.5200, 0.5200, 0.5200),
        (0.4595, 0.4595, 0.4595),
        (0.4100, 0.4100, 0.4100),
        (0.3706, 0.3706, 0.3706),
        (0.2000, 0.2000, 0.2000),
        (     0, 0, 0),
    ]

    # get first N colors 
    np_colors_base = np.array(l_colors_dwt)
    np_colors_rgb = np_colors_base[:num_clusters]

    return np_colors_rgb

def colors_fams_3():
    'custom colors for 3 waves families'

    l_named_colors = [
        'gold', 'darkgreen', 'royalblue',
    ]

    # get rgb colors as numpy array
    np_colors_rgb = np.array(
        [mcolors.to_rgb(c) for c in l_named_colors]
    )

    return np_colors_rgb

def colors_interp(num_clusters):

    # generate spectral colormap
    scm = cm.get_cmap('Spectral', num_clusters)

    # use normalize
    mnorm = mcolors.Normalize(vmin=0, vmax=num_clusters)

    # interpolate colors from cmap
    l_colors = []
    for i in range(num_clusters):
        l_colors.append(scm(mnorm(i)))

    # return numpy array
    np_colors_rgb = np.array(l_colors)[:,:-1]

    return np_colors_rgb


def GetClusterColors(num_clusters):
    'Choose colors or Interpolate custom colormap to number of clusters'

    if num_clusters == 6:
        np_colors_rgb = colors_awt()  # Annual Weather Types

    if num_clusters == 25:
        np_colors_rgb = colors_mjo()  # MJO Categories

    elif num_clusters in [36, 42]:
        np_colors_rgb = colors_dwt(num_clusters)  # Daily Weather Types

    else:
        np_colors_rgb = colors_interp(num_clusters)  # interpolate

    return np_colors_rgb

def GetFamsColors(num_fams):
    'Choose colors or Interpolate custom colormap to number of waves families'

    if num_fams == 3:
        np_colors_rgb = colors_fams_3()  # choosen colors 

    else:
        np_colors_rgb = colors_interp(num_fams)  # interpolate

    return np_colors_rgb



import unittest

import numpy as np
from shapely.geometry import Polygon

from bluemath_tk.core.geo import (
    buffer_area_for_polygon,
    create_polygon,
    filter_points_in_polygon,
    geodesic_azimuth,
    geodesic_distance,
    geodesic_distance_azimuth,
    mask_points_outside_polygon,
    points_in_polygon,
    shoot,
)


class TestGeodesicDistance(unittest.TestCase):
    """Test the geodesic_distance function."""

    def test_scalar_inputs(self):
        """Test geodesic distance calculation with scalar inputs."""

        result = geodesic_distance(0, 0, 0, 90)
        expected = 90.0
        self.assertAlmostEqual(result, expected, delta=0.01)

    def test_array_inputs(self):
        """Test geodesic distance calculation with array inputs."""

        lat1 = np.array([0, 45])
        lon1 = np.array([0, -90])
        lat2 = np.array([0, -45])
        lon2 = np.array([90, 90])
        result = geodesic_distance(lat1, lon1, lat2, lon2)
        expected = np.array([90.0, 180.0])
        np.testing.assert_allclose(result, expected)

    def test_identical_points(self):
        """Test that distance between identical points is zero."""

        result = geodesic_distance(45.0, -120.0, 45.0, -120.0)
        self.assertAlmostEqual(result, 0.0, delta=0.01)

    def test_antipodal_points(self):
        """Test distance between antipodal points (should be 180 degrees)."""

        result = geodesic_distance(0, 0, 0, 180)
        self.assertAlmostEqual(result, 180.0, delta=0.01)


class TestGeodesicAzimuth(unittest.TestCase):
    """Test the geodesic_azimuth function."""

    def test_scalar_inputs(self):
        """Test azimuth calculation with scalar inputs."""

        result = geodesic_azimuth(0, 0, 0, 90)
        expected = 90.0
        self.assertAlmostEqual(result, expected, delta=0.01)

    def test_array_inputs(self):
        """Test azimuth calculation with array inputs."""

        lat1 = np.array([0, 45])
        lon1 = np.array([0, -90])
        lat2 = np.array([0, -45])
        lon2 = np.array([90, 90])
        result = geodesic_azimuth(lat1, lon1, lat2, lon2)
        expected = np.array([90.0, 90.0])
        np.testing.assert_allclose(result, expected)

    def test_north_direction(self):
        """Test azimuth for north direction."""

        result = geodesic_azimuth(0, 0, 90, 0)
        self.assertAlmostEqual(result, 0.0, delta=0.01)

    def test_south_direction(self):
        """Test azimuth for south direction."""

        result = geodesic_azimuth(0, 0, -90, 0)
        self.assertAlmostEqual(result, 180.0, delta=0.01)


class TestGeodesicDistanceAzimuth(unittest.TestCase):
    """Test the geodesic_distance_azimuth function."""

    def test_scalar_inputs(self):
        """Test combined distance and azimuth calculation with scalar inputs."""

        dist, az = geodesic_distance_azimuth(0, 0, 0, 90)
        expected_dist = 90.0
        expected_az = 90.0
        self.assertAlmostEqual(dist, expected_dist, delta=0.01)
        self.assertAlmostEqual(az, expected_az, delta=0.01)

    def test_array_inputs(self):
        """Test combined distance and azimuth calculation with array inputs."""

        lat1 = np.array([0, 45])
        lon1 = np.array([0, -90])
        lat2 = np.array([0, -45])
        lon2 = np.array([90, 90])
        dist, az = geodesic_distance_azimuth(lat1, lon1, lat2, lon2)
        expected_dist = np.array([90.0, 180.0])
        expected_az = np.array([90.0, 90.0])
        np.testing.assert_allclose(dist, expected_dist)
        np.testing.assert_allclose(az, expected_az)


class TestShoot(unittest.TestCase):
    """Test the shoot function."""

    def test_scalar_inputs(self):
        """Test shooting with scalar inputs."""

        lon_f, lat_f, baz = shoot(0, 0, 90, 111.195)  # ~1 degree at equator
        self.assertAlmostEqual(lon_f, 1.0, delta=0.01)
        self.assertAlmostEqual(lat_f, 0.0, delta=0.01)
        self.assertAlmostEqual(baz, 270.0, delta=0.01)

    def test_array_inputs(self):
        """Test shooting with array inputs."""

        lon1 = np.array([0, 0])
        lat1 = np.array([0, 45])
        azimuth = np.array([90, 45])
        maxdist = np.array([111.195, 111.195])
        lon_f, lat_f, baz = shoot(lon1, lat1, azimuth, maxdist)
        self.assertEqual(len(lon_f), 2)
        self.assertEqual(len(lat_f), 2)
        self.assertEqual(len(baz), 2)

    def test_pole_error(self):
        """Test that shooting from pole in non-meridian direction raises error."""

        with self.assertRaises(ValueError):
            shoot(0, 90, 90, 100)  # Shooting east from North Pole
        with self.assertRaises(ValueError):  # Shooting west from South Pole
            shoot(0, -90, 90, 100)  # Shooting west from South Pole


class TestPolygonFunctions(unittest.TestCase):
    """Test polygon-related functions."""

    def setUp(self):
        """Set up test polygon."""

        self.coords = [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]  # Square
        self.polygon = create_polygon(self.coords)

    def test_create_polygon(self):
        """Test polygon creation."""

        poly = create_polygon(self.coords)
        self.assertIsInstance(poly, Polygon)
        self.assertTrue(poly.is_valid)

    def test_points_in_polygon(self):
        """Test point-in-polygon checking."""

        lon = [0.5, 2.0]
        lat = [0.5, 2.0]
        mask = points_in_polygon(lon, lat, self.polygon)
        expected = np.array([True, False])
        np.testing.assert_array_equal(mask, expected)

    def test_points_in_polygon_array_input(self):
        """Test point-in-polygon with numpy array inputs."""

        lon = np.array([0.5, 2.0])
        lat = np.array([0.5, 2.0])
        mask = points_in_polygon(lon, lat, self.polygon)
        expected = np.array([True, False])
        np.testing.assert_array_equal(mask, expected)

    def test_filter_points_in_polygon(self):
        """Test filtering points to keep only those inside polygon."""

        lon = [0.5, 2.0]
        lat = [0.5, 2.0]
        filtered_lon, filtered_lat = filter_points_in_polygon(lon, lat, self.polygon)
        np.testing.assert_array_equal(filtered_lon, np.array([0.5]))
        np.testing.assert_array_equal(filtered_lat, np.array([0.5]))

    def test_points_in_polygon_shape_mismatch(self):
        """Test that mismatched shapes raise ValueError."""

        lon = [0.5, 2.0]
        lat = [0.5]  # Different length
        with self.assertRaises(ValueError):
            points_in_polygon(lon, lat, self.polygon)

    def test_buffer_area_for_polygon(self):
        """Test polygon buffering."""

        buffered = buffer_area_for_polygon(self.polygon, 0.1)
        self.assertIsInstance(buffered, Polygon)
        self.assertGreater(
            buffered.area, self.polygon.area
        )  # Buffered should be larger


class TestMaskPointsOutsidePolygon(unittest.TestCase):
    """Test the mask_points_outside_polygon function."""

    def test_triangle_elements(self):
        """Test masking of triangle elements."""

        elements = np.array([[0, 1, 2], [1, 2, 3]])
        node_coords = np.array([[0, 0], [1, 0], [0, 1], [1, 1]])
        poly = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        mask = mask_points_outside_polygon(elements, node_coords, poly)
        expected = np.array([True, True])  # Both triangles should be inside
        np.testing.assert_array_equal(mask, expected)

    def test_triangles_partially_outside(self):
        """Test triangles that are partially outside the polygon."""

        elements = np.array([[0, 1, 2], [1, 2, 3]])
        node_coords = np.array([[0, 0], [1, 0], [0, 1], [2, 2]])  # Last point outside
        poly = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        mask = mask_points_outside_polygon(elements, node_coords, poly)
        expected = np.array([True, True])
        np.testing.assert_array_equal(mask, expected)


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy.stats import gaussian_kde, probplot
from sklearn.metrics import mean_squared_error

from .base_plotting import DefaultStaticPlotting
from .colors import default_colors


def rmse(pred: np.ndarray, tar: np.ndarray) -> float:
    """
    Calculate the Root Mean Square Error between predicted and target values.

    Parameters
    ----------
    pred : np.ndarray
        Array of predicted values.
    tar : np.ndarray
        Array of target/actual values.

    Returns
    -------
    float
        The Root Mean Square Error value.
    """

    if len(pred) != len(tar):
        raise ValueError("pred and tar must have the same length")

    return np.sqrt(((pred - tar) ** 2).mean())


def bias(pred: np.ndarray, tar: np.ndarray) -> float:
    """
    Calculate the bias between predicted and target values.

    Parameters
    ----------
    pred : np.ndarray
        Array of predicted values.
    tar : np.ndarray
        Array of target/actual values.

    Returns
    -------
    float
        The bias value (mean difference between predictions and targets).
    """

    if len(pred) != len(tar):
        raise ValueError("pred and tar must have the same length")

    return sum(pred - tar) / len(pred)


def si(pred: np.ndarray, tar: np.ndarray) -> float:
    """
    Calculate the Scatter Index between predicted and target values.

    Parameters
    ----------
    pred : np.ndarray
        Array of predicted values.
    tar : np.ndarray
        Array of target/actual values.

    Returns
    -------
    float
        The Scatter Index value.
    """

    if len(pred) != len(tar):
        raise ValueError("pred and tar must have the same length")

    pred_mean = pred.mean()
    tar_mean = tar.mean()

    return np.sqrt(sum(((pred - pred_mean) - (tar - tar_mean)) ** 2) / (sum(tar**2)))


def density_scatter(
    x: np.ndarray, y: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute a density scatter for two arrays using gaussian KDE.

    Parameters
    ----------
    x : np.ndarray
        X values for the scatter plot.
    y : np.ndarray
        Y values for the scatter plot.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        A tuple containing:
        - Sorted x values
        - Sorted y values
        - Density values corresponding to each point
    """

    if len(x) != len(y):
        raise ValueError("x and y must have the same length")

    xy = np.vstack([x, y])
    z = gaussian_kde(xy)(xy)
    idx = z.argsort()
    x1, y1, z = x[idx], y[idx], z[idx]

    return x1, y1, z


def validation_scatter(
    axs: Axes,
    x: np.ndarray,
    y: np.ndarray,
    xlabel: str,
    ylabel: str,
    title: str,
    cmap: str = "rainbow",
) -> None:
    """
    Plot a density scatter and Q-Q plot for validation.

    Parameters
    ----------
    axs : Axes
        Matplotlib axes to plot on.
    x : np.ndarray
        X values for the scatter plot.
    y : np.ndarray
        Y values for the scatter plot.
    xlabel : str
        Label for the X-axis.
    ylabel : str
        Label for the Y-axis.
    title : str
        Title for the plot.
    cmap : str, optional
        Colormap to use for the scatter plot. Default is "rainbow".
    """

    x2, y2, z = density_scatter(x, y)

    # plot
    axs.scatter(x2, y2, c=z, s=5, cmap=cmap)

    # labels
    axs.set_xlabel(xlabel)
    axs.set_ylabel(ylabel)
    axs.set_title(title)

    # axis limits
    maxt = np.ceil(max(max(x) + 0.1, max(y) + 0.1))
    axs.set_xlim(0, maxt)
    axs.set_ylim(0, maxt)
    axs.plot([0, maxt], [0, maxt], "-r")
    axs.set_xticks(np.linspace(0, maxt, 5))
    axs.set_yticks(np.linspace(0, maxt, 5))
    axs.set_aspect("equal")

    # qq-plot
    xq = probplot(x, dist="norm")
    yq = probplot(y, dist="norm")
    axs.plot(xq[0][1], yq[0][1], "o", markersize=0.5, color="k", label="Q-Q plot")

    # diagnostic errors
    props = dict(
        boxstyle="round", facecolor="w", edgecolor="grey", linewidth=0.8, alpha=0.5
    )
    mse = mean_squared_error(x2, y2)
    rmse_e = rmse(x2, y2)
    BIAS = bias(x2, y2)
    SI = si(x2, y2)
    label = "\n".join(
        (
            r"RMSE = %.2f" % (rmse_e,),
            r"MSE =  %.2f" % (mse,),
            r"BIAS = %.2f" % (BIAS,),
            R"SI = %.2f" % (SI,),
        )
    )
    axs.text(
        0.05,
        0.95,
        label,
        transform=axs.transAxes,
        fontsize=9,
        verticalalignment="top",
        bbox=props,
    )


def plot_scatters_in_triangle(
    dataframes: List[pd.DataFrame],
    data_colors: Optional[List[str]] = None,
    **kwargs,
) -> Tuple[Figure, np.ndarray]:
    """
    Plot scatter plots of the dataframes with axes in a triangle arrangement.

    Parameters
    ----------
    dataframes : List[pd.DataFrame]
        List of dataframes to plot. Each dataframe should contain the same columns.
    data_colors : Optional[List[str]], optional
        List of colors for the dataframes. If None, uses default_colors.
    **kwargs : dict, optional
        Additional keyword arguments for the scatter plot. These will be passed to
        matplotlib.pyplot.scatter. Common parameters include:
        - s : float, marker size
        - alpha : float, transparency
        - marker : str, marker style

    Returns
    -------
    Tuple[Figure, np.ndarray]
        A tuple containing:
        - Figure object
        - 2D array of Axes objects

    Raises
    ------
    ValueError
        If the variables in the first dataframe are not present in all other dataframes.
    """

    if data_colors is None:
        data_colors = default_colors

    # Get the number and names of variables from the first dataframe
    variables_names = list(dataframes[0].columns)
    num_variables = len(variables_names)

    # Check variables names are in all dataframes
    for df in dataframes:
        if not all(v in df.columns for v in variables_names):
            raise ValueError(
                f"Variables {variables_names} are not in dataframe {df.columns}."
            )

    # Create figure and axes
    default_static_plot = DefaultStaticPlotting()
    fig, axes = default_static_plot.get_subplots(
        nrows=num_variables - 1,
        ncols=num_variables - 1,
        sharex=False,
        sharey=False,
    )
    if isinstance(axes, Axes):
        axes = np.array([[axes]])

    for c1, v1 in enumerate(variables_names[1:]):
        for c2, v2 in enumerate(variables_names[:-1]):
            for idf, df in enumerate(dataframes):
                default_static_plot.plot_scatter(
                    ax=axes[c2, c1],
                    x=df[v1],
                    y=df[v2],
                    c=data_colors[idf],
                    alpha=0.6,
                    **kwargs,
                )
            if c1 == c2:
                axes[c2, c1].set_xlabel(variables_names[c1 + 1])
                axes[c2, c1].set_ylabel(variables_names[c2])
            elif c1 > c2:
                axes[c2, c1].xaxis.set_ticklabels([])
                axes[c2, c1].yaxis.set_ticklabels([])
            else:
                fig.delaxes(axes[c2, c1])

    return fig, axes




import os.path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import xarray as xr
import pickle
import sys

#----------------------------
Scenarios = ['0.3', '0.5', '1.0', '1.5']#, '2.0', '2.5']
Subscenario = 'MED'
# Subscenario = 'HIGH'
# Subscenario = 'LOW'

#----------------------------
# # Kwajalein
lon = 167.5
lat = 9.75

# Guam
#lon = 144.5
#lat = 13.5

# # San Diego
# lon = -117.17
# lat = 32.71

# MAJURO
#lon = 171.3725
#lat = 7.1060

#----------------------------
# location = 'GMSL'
location = 'TGs'
#location = 'RegionalGrid'


#------------------------------------------------------------------------
# 1) read excel data and select site
# %%
rutin = '/Users/anacrueda/Documents/Proyectos/SERDP/niveles_ref_proyecc/'
file = 'techrpt083.csv'

data = pd.read_csv(os.path.join(rutin, file), skiprows=15)


if location == 'GMSL': # Global SL

    data = data.iloc[:18]

else: # Regional SL

    if location == 'TGs':
        data = data.iloc[18:4302]

    elif location == 'RegionalGrid':
        data = data.iloc[4302:]
# %%
#print(data)

# find closest point:
dif_lon = data['Longitude'] - lon
dif_lat = data['Latitude'] - lat
dif = np.sqrt(dif_lon**2 + dif_lat**2)

min_dist = dif.min(skipna=True)
min_ind = dif.idxmin(skipna=True)


site = data['Site'].loc[min_ind]
data_site = data.loc[(data['Site'] == site)]

# %%
#print(data_site.T)
#print()


#------------------------------------------------------------------------
# 2) Fit RSL to a 2º order polynomial
# based on matlab code that Peter's group used to interpolate the Sweet et al. (2017) SLR decadal projection to hourly data


time_axis = np.array(['2000-01-01', '2010-01-01', '2020-01-01', '2030-01-01', '2040-01-01', '2050-01-01', '2060-01-01',
                    '2070-01-01', '2080-01-01', '2090-01-01', '2100-01-01', '2120-01-01', '2150-01-01', '2200-01-01'], dtype='datetime64[h]')
time_axis_rel = (time_axis - time_axis[0]).astype('int')


time_axis_H = np.arange('2015-01-01', '2101-01-01', dtype='datetime64[h]')
time_axis_H_rel = (time_axis_H - time_axis[0]).astype('int')

y_slr = []
fig, ax = plt.subplots(1,1, figsize=(12, 9))
for scenario in Scenarios:

    data_subset = data_site.loc[data_site['Scenario'] == (scenario + ' - ' + Subscenario)]

    y = data_subset.to_numpy()[0][6:]
    y = y.astype('int')
    y = y/100.0 # to meters

    # Fit data, from 2000 to 2100, to a 2º polynomial function
    coef = np.polyfit(time_axis_rel[:-3], y[:-3], 2)

    # Obtain hourly data from the adjusted polynomial function
    y_fit_H = coef[0]*time_axis_H_rel**2 + coef[1]*time_axis_H_rel + coef[2]

    y_slr.append(y_fit_H)
    # plot
    ax.plot(time_axis, y, '.', label='')
    ax.plot(time_axis_H, y_fit_H, label=scenario)

ax.set_xlim([time_axis[0], time_axis[-3]])

ax.set_ylim([0, 2])
ax.set_title(site + ', SL rise', fontweight='bold')
ax.set_ylabel('m')
ax.grid()
plt.legend(title='GMSL rise scenarios')
# plt.show()
fig.savefig(os.path.join(rutin, 'SLR_' + site + '.png'))

# %%

print(y_slr)
print(np.shape(y_slr))
print()
print(Scenarios)

# %%
xds_slr = xr.Dataset(
    {'slr': (('scenario', 'time'), y_slr),
},
    coords={'scenario': Scenarios, 'time': time_axis_H}
)

print(xds_slr)

# %%
file_Name =r'/Users/anacrueda/Documents/Proyectos/SERDP/niveles_ref_proyecc/slr_kwa.pkl'
# open the file for writing
fileO = open(file_Name,'wb')

# this writes the object a to the
# file named 'testfile'
pickle.dump(xds_slr,fileO)

# here we close the fileObject
fileO.close()

# %%
#file_csv = r'/Users/anacrueda/Documents/Proyectos/SERDP/niveles_ref_proyecc/slr_majuro_05.csv'

#slr_a = xds_slr.to_array()
#slr_c = slr_a.sel(scenario='0.5').values[:].flatten()
#.to_array()

#print(np.shape(slr_c))


#slr_d = xr.Dataset(
#    {'slr': ('time', slr_c),
#},
#    coords={'time': slr_a.time.values[:]}
#)

#print(slr_d)

#slr_x = slr_d.to_array()

# %%
#slr_b = slr_x.to_pandas().T
#print(slr_b)
# %%
#slr_b.to_csv(file_csv)
#print(slr_b)

#xds_slr.to_csv(file_csv)
